
(define vendor "SYNOPSYS")
(define program "PHYSICAL SYNTHESIS")
(define version "")
(define preRouteFormat 2) ; 0 - place obs, 1 - routing obs, 2 - pnet
(define designLevel 1) ; 0 - floorplan, 1 - floorplan + placement
                       ; 2 - floorplan + placement + complete routing
(define debugFlag #f)
(define pdefBusNameRule "[%d]")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpPDEF3
;  Function: Dump the design into a PDEF file
;  Last Modifed: 8/8/2000
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpPDEF3 (lambda (site fileName)
  (let*
    (
      (cv    (geGetEditCell))
      (lib   (dbGetLibIdByCellId cv))
      (file  (open-output-file fileName))
      (name  (dbFetchObjectField cv '() "name"))
      (bbox    '())
      (area      0)
      (fp      '())
      (logfile '())
    )

    ; Initilization
    (define viaArrayMasterList (make-vector 20 '()))
    (define viaMasterList '())
    (define contactCodeIndexList '())
    (define pnetList '())
    (define netList '())
    (define wireMasterList (make-vector 1000 0))
    (define horizontalWireList '())
    (define verticalWireList '())

    (define dbResolution
      (dbFetchObjectField lib '() "dataBaseUnitPerUserDistUnit")
    )
    (db-foreach cv '() "WireMaster" master
      (begin
        (vector-set! wireMasterList
          (dbFetchObjectField cv master "index")
          (/ (dbFetchObjectField cv master "width") dbResolution)
        )
        #t
      )
    )
    (db-foreach cv '() "HorizontalWire" wire
      (begin
        (set! horizontalWireList (cons wire horizontalWireList))
        #t
      )
    )
    (db-foreach cv '() "VerticalWire" wire
      (begin
        (set! verticalWireList (cons wire verticalWireList))
        #t
      )
    )

    (display "( CLUSTERFILE" file)
    (newline file)
    (display "  ( PDEFVERSION \"3.0\" )" file)
    (newline file)
    (display "  ( DESIGN \"" file)
    (display name file)
    (display "\" )" file)
    (newline file)
    (display "  ( DATE \"\" )" file)
    (newline file)
    (display "  ( VENDOR \"" file)
    (display vendor file)
    (display "\" )" file)
    (newline file)
    (display "  ( PROGRAM \"" file)
    (display program file)
    (display "\" )" file)
    (newline file)
    (display "  ( VERSION \"" file)
    (display version file)
    (display "\" )" file)
    (newline file)
    (display "  ( DIVIDER / )" file)
    (newline file)
    (display "  ( PIN_DELIMITER / )" file)
    (newline file)
    (display "  ( BUS_DELIMITER [ ] )" file)
    (newline file)
    (display "  ( NETLIST_TYPE VERILOG )" file)
    (newline file)
    (display "  ( DESIGN_FLOW \"\" )" file)
    (newline file)
    (display "  ( DISTANCE_UNIT " file)
    (display (/ 1.0 dbResolution) file)
    (display " )" file)
    (newline file)
    (display "  ( DEF_CONVERSION_FACTOR " file)
    (display dbResolution file)
    (display " )" file)
    (newline file)

    (if (equal? #t debugFlag)
      (dumpFP ".dumpFP.log")
    )
    (display "Translating design into IEEE PDEF..........")(newline)
    (display "    Site array information...")(newline)
    (dumpCellRow cv site file)

    ; Dump core area
    (db-foreach cv '() "BaseArray" base
      (begin
        (set! bbox (dbFetchObjectField cv base "bbox"))
        #f ; Assume there is only one base array - untrue for rectilinear design
      )
    )
    (display "  ( CORE_AREA " file)
    (display (* dbResolution (caar bbox)) file)
    (display " " file)
    (display (* dbResolution (cadar bbox)) file)
    (display " " file)
    (display (* dbResolution (caadr bbox)) file)
    (display " " file)
    (display (* dbResolution (cadadr bbox)) file)
    (display " )" file)
    (newline file)

    (display "    Routing track information...")(newline)
    (dumpTrack cv file)
    (display "    Routing layer information...")(newline)
    (dumpLayer cv file)
    (display "    Via information...")(newline)
    (dumpVia cv file)

    (display "  ( CLUSTER FP_CLUSTER" file)
    (newline file)

    (display "    IO pin/pad location information...")(newline)
    (dumpIOPin cv file)
    (dumpIOPadPlace cv file)
    (display "    Macro location information...")(newline)
    (dumpMacroPlace cv file)
    (cond
      ((= designLevel 0)
        (display "    Fixed standard cell location information...")(newline)
        (dumpFixedCellPlace cv file)
      )
      ((> designLevel 0)
        (display "    Standard cell placement information...")(newline)
        (dumpCellPlace cv file)
      )
    )

    (display "    Region information...")(newline)
    (dumpRegion cv file)
    (display "    Placement obstruction information...")(newline)
    (dumpPlaceBlockage cv file)

    (if (not (= designLevel 2))
      (begin
        (display "    Pre-route information...")(newline)
        (if (= preRouteFormat 2)
          (dumpPNET cv file)
          (dumpPreRouteObs cv file preRouteFormat)
        )
      )
      (begin
        (display "    Detail route information...")(newline)
        (dumpPNET cv file)
        (dumpRoute cv file)
      )
    )
    (display "  )" file)
    (newline file)
    (display ")" file)
    (newline file)
    (close-output-port file)

    (if (equal? #t debugFlag)
      (begin
        (set! fp (open-input-file ".dumpFP.log"))
        (set! logfile (open-output-file "dumpPDEF3.log"))
        (add-dumpFP-log fp logfile)
        (newline logfile)
        (close-input-port fp)
        (close-output-port logfile)
        (system "/bin/rm -f .dumpFP.log")
      )
    )

    (display "Translation is complete.") (newline)
    #t
  )
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpCellRow
;  Function: Dump the cell rows info into a file in PDEF 3.0 format
;  Last Modifed: 5/22/2000
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpCellRow (lambda (cv site file)
  (let*
    (
      (lib  (dbGetLibIdByCellId cv))
      (baseArrayId        0)
      (tileHeight         0)
      (tileWidth          0)
      (bBox             '())
      (rowWidth           0)
      (numOfTile          0)
      (rowDir             0)
      (dir                "")
      (flip               0)
      (orient             0)
      (index              0)
    )
    (db-foreach cv '() "CellRow" row
      (begin
        (set! baseArrayId (dbFetchObjectField cv row "baseArrayId"))
        (set! tileWidth (dbFetchObjectField cv baseArrayId "tileWidth"))
        (set! tileWidth (/ tileWidth dbResolution))
        (set! bBox (dbFetchObjectField cv row "bbox"))
        (set! rowDir (dbFetchObjectField cv row "orientation"))
        (if (= rowDir 0)
	  (begin
            (set! dir "H")
	    (set! rowWidth (- (caadr bBox) (caar bBox)))
	  )
	  (begin
            (set! dir "V")
	    (set! rowWidth (- (cadadr bBox) (cadar bBox)))
	  )
        )
	(set! numOfTile (/ rowWidth tileWidth))
        (set! flip (dbFetchObjectField cv row "tile"))
        (if (= flip 0)
          (set! orient "0")
          (set! orient "180-mirror")
        )
        (display "  ( SITE " file)
        (display "ROW_" file)
        (display index file)
        (display " " file)
        (display site file)
        (display " \"" file)
        (display (* dbResolution (caar bBox)) file)
        (display " " file)
        (display (* dbResolution (cadar bBox)) file)
        (display "\" \"" file)
        (display dir file)
        (display "\" \"" file)
        (display orient file)
        (display "\" \"" file)
        (display (* dbResolution tileWidth) file)
        (display "\" \"" file)
        (display numOfTile file)
        (display "\" )" file)
        (newline file)
        (set! index (+ index 1))
        #t
      ) ; end begin
    ) ; end db-foreach
  ) ; end let*
)) ; end lambda and define

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpTrack
;  Function: Dump the wire tracks information into PDEF
;  Last Modifed: 5/22/2000
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpTrack (lambda (cv file)
  (let*
    (
      (lib  (dbGetLibIdByCellId cv))
      (bbox     '())
      (layer      0)
      (layerName "")
      (dir       "")
      (direction "")
      (start      0)
      (step       0)
      (trackCount 0)
      (index      0)
    )
    (set! bbox (dbFetchObjectField cv '() "cellBBox"))

    (db-foreach cv '() "WireTrack" track
      (begin
        (set! layer (dbFetchObjectField cv track "layer"))
        (set! layerName (dbGetLayerNameFromNumber lib layer))
        (set! dir (dbFetchObjectField cv track "trackDirection"))
        (if (= 0 dir)
          (set! direction "Y") ; horizontal track
          (set! direction "X") ; vertical track
        )
        (set! trackCount (dbFetchObjectField cv track "times"))
        (set! start
          (dbFetchObjectField cv track "startTrack")
        )
        (set! step
          (dbFetchObjectField cv track "trackStep")
        )

        (display "  ( DEF_TRACK_" file)
        (display index file)
        (display " \"" file)
        (display direction file)
        (display "\" \"" file)
        (display start file)
        (display "\" \"" file)
        (display trackCount file)
        (display "\" \"" file)
        (display step file)
        (display "\" \"" file)
        (display layerName file)
        (display "\")" file)
        (newline file)
        (set! index (+ index 1))
      #t
      )
    ) ; end db-foreach

    (display "  ( DEF_DIEAREA " file)
    (display "\"" file)
    (display (* dbResolution (caar bbox)) file)
    (display "\" \"" file)
    (display (* dbResolution (cadar bbox)) file)
    (display "\" \"" file)
    (display (* dbResolution (caadr bbox)) file)
    (display "\" \"" file)
    (display (* dbResolution (cadadr bbox)) file)
    (display "\" )" file)
    (newline file)
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpLayer
;  Function: Dump the metal layer numbers into a file in PDEF 3.0
;  Last Modifed: 7/4/2000
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpLayer (lambda (cv file)
  (let*
    (
      (layer  0)
      (lib (dbGetLibIdByCellId cv))
      (layerList (dbFetchLayerIdList lib))
      (len (length layerList))
      (maskName "")
      (layerName "")
      (maskList (list "metal1" "metal2" "metal3" "metal4" "metal5" "metal6"
        "metal7" "via1" "via2" "via3" "via4" "via5" "via6" "poly" "polyCont")
      )
      (hList '("metal1" "metal3" "metal5" "metal7"))
      (vList '("poly" "metal2" "metal4" "metal6"))
      (dir #f)
    )
    (display "  ( LAYER_DEF" file)
    (newline file)
    (do
      ((i 0 (+ i 1)))
      ((equal? i len))
      (set! layer (list-ref layerList i))
      (set! maskName (dbFetchLayerInfo lib layer "maskName"))
      (if (not (equal? maskName ""))
        (begin
          (set! layerName (dbGetLayerNameFromNumber lib layer))
          (if (equal? layerName #f)
            (begin
              (display "Error: There is no layer name defined for layer ")
              (display layer)
              (newline)
            )
            (if (member maskName maskList)
              (begin
                (display "    ( LAYER " file)
                (display layerName file)
                (display " " file)
                (display layer file)
                (display " )" file)
                (newline file)
              )
            ) ; end if (member ...
          ) ; end if (equal? ...
        ) ; end begin
      ) ; end if (not (equal? ...
      #t
    ) ; end do
    (display "  )" file)
    (newline file)
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpVia
;  Function: Dump the via information into PDEF
;  Last Modifed: 9/3/2000
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpVia (lambda (cv file)
  (let*
    (
      (lib           (dbGetLibIdByCellId cv))
      (viaList      '())
      (ul             0)
      (ll             0)
      (cut            0)
      (cutName       "")
      (ulName        "")
      (llName        "")
      (width          0)
      (height         0)
      (ulEncWidth     0)
      (ulEncHeight    0)
      (llEncWidth     0)
      (llEncHeight    0)
      (lx             0)
      (ly             0)
      (ux             0)
      (uy             0)
      (index          0)
      (contactCodeIdx 0)
    )
    (set! viaList (dbFetchContactNames))
    (display "  ( VIA_DEF" file)
    (newline file)
    (for-each ( lambda (via)
      (begin
        (set! index (+ index 1))
        (set! contactCodeIdx (dbFetchNonDefaultContactCodeInfo lib via "contactCodeNumber"))
        (set! cut (dbFetchNonDefaultContactCodeInfo lib via "cutLayerNumber"))
        (set! ul (dbFetchNonDefaultContactCodeInfo lib via "upperLayerNumber"))
        (set! ll (dbFetchNonDefaultContactCodeInfo lib via "lowerLayerNumber"))
        (set! cutName (dbGetLayerNameFromNumber lib cut))
        (set! ulName (dbGetLayerNameFromNumber lib ul))
        (set! llName (dbGetLayerNameFromNumber lib ll))
        (set! width (dbFetchNonDefaultContactCodeInfo lib via "cutWidth"))
        (set! height (dbFetchNonDefaultContactCodeInfo lib via "cutHeight"))
        (set! ulEncWidth (dbFetchNonDefaultContactCodeInfo lib via "upperLayerEncWidth"))
        (set! ulEncHeight (dbFetchNonDefaultContactCodeInfo lib via "upperLayerEncHeight"))
        (set! llEncWidth (dbFetchNonDefaultContactCodeInfo lib via "lowerLayerEncWidth"))
        (set! llEncHeight (dbFetchNonDefaultContactCodeInfo lib via "lowerLayerEncHeight"))
        (set! viaMasterList
          (cons (list contactCodeIdx cutName ulName llName) viaMasterList)
        )
        (display "    ( VIA " file)
        (display via file)
        (display " " file)
        (display index file)
        (newline file)

        (display "      ( VIA_RECT " file)
        (display ulName file)
        (display " \" " file)
        (set! lx (- (- 0 (/ width 2)) ulEncWidth))
        (set! ly (- (- 0 (/ height 2)) ulEncHeight))
        (set! ux (+ (/ width 2) ulEncWidth))
        (set! uy (+ (/ height 2) ulEncHeight))
        (display lx file)(display " " file)
        (display ly file)(display " " file)
        (display ux file)(display " " file)
        (display uy file)(display " \" )" file)
        (newline file)

        (display "      ( VIA_RECT " file)
        (display cutName file)
        (display " \" " file)
        (set! lx (- 0 (/ width 2)))
        (set! ly (- 0 (/ height 2)))
        (set! ux (/ width 2))
        (set! uy (/ width 2))
        (display lx file)(display " " file)
        (display ly file)(display " " file)
        (display ux file)(display " " file)
        (display uy file)(display " \" )" file)
        (newline file)

        (display "      ( VIA_RECT " file)
        (display llName file)
        (display " \" " file)
        (set! lx (- (- 0 (/ width 2)) llEncWidth))
        (set! ly (- (- 0 (/ height 2)) llEncHeight))
        (set! ux (+ (/ width 2) llEncWidth))
        (set! uy (+ (/ height 2) llEncHeight))
        (display lx file)(display " " file)
        (display ly file)(display " " file)
        (display ux file)(display " " file)
        (display uy file)(display " \" )" file)
        (newline file)

        (display "    )" file)
        (newline file)
      )) ; end begin and lambda
      viaList
    ) ; end for-each
    (set! viaMasterList (reverse viaMasterList))
    (display "    Via Array Information ...")(newline)
    (dumpViaArrayMaster cv file index)
    (display " )" file)
    (newline file)
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpViaArrayMaster
;  Function: Dump the via array masters into PDEF
;  Last Modifed: 9/3/2000
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpViaArrayMaster (lambda (cv file index)
  (let*
    (
      (lib              0)
      (contactCodeIdx   0)
      (contactName     "")
      (xDuplicate       0)
      (yDuplicate       0)
      (xPitch           0)
      (yPitch           0)
      (bbox           '())
      (flag            #t)
      (match          '())
      (new            '())
      (pattern         "")
      (viaArrayMaster '())
      (nth            '())
    )

    (set! lib (dbGetLibIdByCellId cv))
    (db-foreach cv '() "ContactArray" obj
      (begin
        (set! contactCodeIdx (dbFetchObjectField cv obj "contactCodeIdx"))
        (set! contactName (getContactNameByContactCodeIdx lib contactCodeIdx))
        (set! xDuplicate (dbFetchObjectField cv obj "xDuplicate"))
        (set! yDuplicate (dbFetchObjectField cv obj "yDuplicate"))
        (set! xPitch (dbFetchObjectField cv obj "xPitch"))
        (set! yPitch (dbFetchObjectField cv obj "yPitch"))
        (set! nth '())
        (call-with-current-continuation (lambda (break)
          (do
            ((i 0 (+ i 1)))
            ((= i (length contactCodeIndexList)))
            (if (= (list-ref contactCodeIndexList i) contactCodeIdx)
              (begin
                (set! nth i)
                (break nth)
              )
            )
          ) ;end do
        )) ;end call-with-current-continuation
        (if (null? nth)
          (begin
            (set! nth (length contactCodeIndexList))
            (set! contactCodeIndexList
              (append contactCodeIndexList (list contactCodeIdx))
            )
          )
        )
        (set! viaArrayMaster (vector-ref viaArrayMasterList nth))

        (call-with-current-continuation (lambda (ret)
          (for-each ( lambda (master)
            (begin
              (set! flag #t)
              (call-with-current-continuation (lambda (cont)
                (if (not (= (car master) contactCodeIdx))
                  (begin
                    (set! flag #f)
                    (cont flag)
                  )
                )
                (if (not (= (cadr master) xDuplicate))
                  (begin
                    (set! flag #f)
                    (cont flag)
                  )
                )
                (if (not (= (caddr master) yDuplicate))
                  (begin
                    (set! flag #f)
                    (cont flag)
                  )
                )
                (if (not (= (cadddr master) xPitch))
                  (begin
                    (set! flag #f)
                    (cont flag)
                  )
                )
                (if (not (= (car (cddddr master)) yPitch))
                  (begin
                    (set! flag #f)
                    (cont flag)
                  )
                )
              )) ; end lambda and call-with-current-continuation

              (if (equal? #t flag)
                (begin
                  (set! match master)
                  (ret match)
                )
                (set! match '())
              )
            )) ; end begin and lambda
            viaArrayMaster
          ) ; end for-each
        )) ; end lambda and call-with-current-continuation

        (if (null? match)
          (begin
            (set! index (+ index 1))
            (set! pattern (list->string
              (list contactName "_"
                (number->string xDuplicate) "_"
                (number->string yDuplicate) "_"
                (number->string xPitch)  "_"
                (number->string yPitch)
              )
            ))
            (set! new (list contactCodeIdx xDuplicate yDuplicate
              xPitch yPitch pattern index obj))
            (set! viaArrayMaster (append viaArrayMaster (list new)))
            (vector-set! viaArrayMasterList nth viaArrayMaster)
          ) ; end begin
        ) ; end if
        (set! match '())
        #t
      ) ; end begin
    ) ; end db-foreach
    (viaPatternGenerate cv file)
    #t
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: viaPatternGenerate
;  Function: Generate via geometry pattern for contact array master
;  Last Modifed: 6/8/2000
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define viaPatternGenerate (lambda (cv file)
  (let*
    (
      (lib             (dbGetLibIdByCellId cv))
      (origin        '())
      (contId          0)
      (geomList      '())
      (layerName      "")
      (x0              0)
      (x1              0)
      (y0              0)
      (y1              0)
      (masterList    '())
    )
    (do
      ((i 0 (+ i 1)))
      ((= i (length contactCodeIndexList)))
      (set! masterList (vector-ref viaArrayMasterList i))
      (for-each (lambda (via)
        (begin
          (set! contId (car (reverse via)))
          (set! origin (car (dbFetchObjectField cv contId "position")))
          (set! geomList (dbFetchContactGeom cv contId))
          (display "    ( VIA " file)
          (display (list-ref via 5) file)
          (display " " file)
          (display (list-ref via 6) file)
          (newline file)

          (for-each (lambda (geom)
            (begin
              (set! layerName (dbGetLayerNameFromNumber lib (car geom)))
              (set! x0 (round (* (- (car (caadr geom)) (car origin)) dbResolution)))
              (set! x1 (round (* (- (car (cadadr geom)) (car origin)) dbResolution)))
              (set! y0 (round (* (- (cadr (caadr geom)) (cadr origin)) dbResolution)))
              (set! y1 (round (* (- (cadr (cadadr geom)) (cadr origin)) dbResolution)))

              (display "      ( VIA_RECT " file)
              (display layerName file)
              (display " \" " file)
              (display x0 file)(display " " file)
              (display y0 file)(display " " file)
              (display x1 file)(display " " file)
              (display y1 file)(display " \" )" file)
              (newline file)
            )) ; end begin and lambda
            geomList
          ) ; end for-each
          (display "    )" file)
          (newline file)
        )) ; end begin and lambda
        masterList
      ) ; end for-each
    ) ; end do
  )
))

(define getContactNameByContactCodeIdx (lambda (lib index)
  (let*
    (
      (contactNameList (dbFetchContactNames))
      (ret '())
    )
    (for-each (lambda (cont)
      (if (equal? index (dbFetchNonDefaultContactCodeInfo lib cont "contactCodeNumber"))
        (set! ret (dbFetchNonDefaultContactCodeInfo lib cont "name"))
      ))
      contactNameList
    )
    ret
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpIOPin
;  Function: Dump the IO locations into PDEF
;  Last Modifed: 8/20/2000
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpIOPin (lambda (cv file)
  (let*
    (
      (lib       (dbGetLibIdByCellId cv))
      (pinName     "")
      (pinNameList '())
      (pinNumber   '())
      (layer       "")
      (layerName   "")
      (type        0)
      (pin_type    "")
      (netId       0)
      (netName     "")
      (dir         0)
      (pin_dir     "")
      (bbox       '())
      (x_size      0)
      (y_size      0)
    )
    (db-foreach cv '() "Pin" pin
      (begin
        (set! pinName (dbFetchObjectField cv pin "name"))
        (set! pinNameList (string-parse pinName))
        (set! pinName (car pinNameList))
        (if (> (length pinNameList) 1)
          (set! pinNumber (cadr pinNameList))
          (set! pinNumber '())
        )
        (set! pinName (busNameRule pinName))
        (set! layer (dbFetchObjectField cv pin "layer"))
        (set! layerName (dbGetLayerNameFromNumber lib layer))
        (set! dir (dbFetchObjectField cv pin "accessDirection"))
        (set! bbox (dbFetchObjectField cv pin "bbox"))
        (set! x_size (- (caadr bbox) (caar bbox)))
        (set! y_size (- (cadadr bbox) (cadar bbox)))
        (set! netId (dbFetchObjectField cv pin "netId"))
        (set! netName (dbFetchObjectField cv netId "name"))
        (set! type (dbFetchObjectField cv netId "type"))
        (cond
          ((> (bit-and dir dbbPinInputPort) 0)
            (set! pin_dir "INPUT")
          )
          ((> (bit-and dir dbbPinOutputPort) 0)
            (set! pin_dir "OUTPUT")
          )
          ((> (bit-and dir dbbPinInputOutputPort) 0)
            (set! pin_dir "INOUT")
          )
          ((> (bit-and dir dbbPinTriState) 0)
            (set! pin_dir "OUTPUT TRISTATE")
          )
          (else
	    (set! pin_dir "INOUT")
          )
        ) ; end cond
        (cond
          ((> (bit-and type dbbSignalNet) 0)
            (set! pin_type "SIGNAL")
          )
          ((> (bit-and type dbbPowerNet) 0)
            (set! pin_type "POWER")
          )
          ((> (bit-and type dbbGroundNet) 0)
            (set! pin_type "GROUND")
          )
          ((> (bit-and type dbbClockNet) 0)
            (set! pin_type "CLOCK")
          )
          (else
	    (set! pin_type "SIGNAL")
          )
        ) ; end cond

        (display "    ( PIN " file)
        (display pinName file)
        (newline file)
        (if (not (null? pinNumber))
          (begin
            (display "      ( PIN_NUMBER " file)
            (display pinNumber file)
            (display " )" file)
            (newline file)
          )
        )
        (display "      ( LOC " file)
        (display (* dbResolution (caar bbox)) file)
        (display " " file)
        (display (* dbResolution (cadar bbox)) file)
        (display " )" file)
        (newline file)
        (display "      ( ORIENT \"0\" )" file)
        (newline file)
        (display "      ( PLACED )" file)
        (newline file)
        (display "      ( PIN_TYPE " file)
        (display pin_type file)
        (display " )" file)
        (newline file)
        (display "      ( PIN_DIR " file)
        (display pin_dir file)
        (display " )" file)
        (newline file)
        (display "      ( PIN_LAYER \"" file)
        (display layerName file)
        (display " ( " file)
        (display  0 file)
        (display " " file)
        (display  0 file)
        (display " ) ( " file)
        (display (round (* dbResolution x_size)) file)
        (display " " file)
        (display (round (* dbResolution y_size)) file)
        (display " )\" )" file)
        (newline file)
        (display "      ( DEF_NET_NAME \"" file)
        (display netName file)
        (display "\" )" file)
        (newline file)
        (display "    )" file)
        (newline file)
        #t
      ) ; end begin
    ) ; end db-foreach
    #t
  ) ; end let*
)) ; end lambda and define

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpIOPadPlace
;  Function: Dump IO pads (but not pad fillers) into a file
;            in PDEF 3.0 format
;  Last Modifed: 9/6/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpIOPadPlace (lambda (cv file)
  (let*
    (
      (flag        0)
      (bBox      '())
      (transform  "")
      (orient     "")
      (gateName   "")
      (x0          0)
      (y0          0)
      (info      '())
    )
    (db-foreach cv '() "cellInstMaster" master
      (begin
        (set! flag (dbFetchObjectField cv master "cellFlag"))
        (set! gateName (dbFetchObjectField cv master "cellName"))
        (if (isPadCell flag)
        (begin
          (db-foreach cv master "CellInst" cell
            (begin
              (set! info (getCellInstLocOrient cv cell master))
              (set! x0 (car info))
              (set! y0 (cadr info))
              (set! orient (caddr info))

              (display "    ( CELL " file)
              (display (dbFetchObjectField cv cell "name") file)
              (newline file)
              (display "      ( GATE_NAME " file)
              (display gateName file)
              (display " )" file)
              (newline file)
              (display "      ( LOC " file)
              (display (* dbResolution x0) file)
              (display " " file)
              (display (* dbResolution y0) file)
              (display " )" file)
              (newline file)
              (display "      ( RESTRICTION \"FIXED_PLACEMENT\" )" file)
              (newline file)
              (display "      ( ORIENT " file)
              (display "\"" file)
              (display orient file)
              (display "\" )" file)
              (newline file)
              (display "      ( FIXED )" file)
              (newline file)
              (display "    )" file)
              (newline file)
              #t
            ) ; end begin
          ) ; end db-foreach
        )) ; end begin and if
        #t
      ) ; end begin
    ) ; end db-foreach
  ) ; end let*
)) ; end lambda and define

(define isPadCell (lambda (flag)
  (and
    (or
      (> (bit-and flag dbbIOPadCell) 0)
      (> (bit-and flag dbbCornerPadCell) 0)
    )
    (= (bit-and flag dbbFillerCell) 0) ; filter out pad fillers
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpMacroPlace
;  Function: Dump the macro locations into a file in PDEF 3.0 format
;  Last Modifed: 9/6/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpMacroPlace (lambda (cv file)
  (let*
    (
      (flag        0)
      (bBox      '())
      (transform  "")
      (orient     "")
      (gateName   "")
      (x0          0)
      (y0          0)
      (info      '())
    )
    (db-foreach cv '() "cellInstMaster" master
      (begin
        (set! flag (dbFetchObjectField cv master "cellFlag"))
        (set! gateName (dbFetchObjectField cv master "cellName"))
        (if (> (bit-and flag dbbMacroCell) 0)
        (begin
          (db-foreach cv master "CellInst" cell
            (begin
              (set! info (getCellInstLocOrient cv cell master))
              (set! x0 (car info))
              (set! y0 (cadr info))
              (set! orient (caddr info))

              (display "    ( CELL " file)
              (display (dbFetchObjectField cv cell "name") file)
              (newline file)
              (display "      ( GATE_NAME " file)
              (display gateName file)
              (display " )" file)
              (newline file)
              (display "      ( LOC " file)
              (display (* dbResolution x0) file)
              (display " " file)
              (display (* dbResolution y0) file)
              (display " )" file)
              (newline file)
              (display "      ( RESTRICTION \"FIXED_PLACEMENT\" )" file)
              (newline file)
              (display "      ( ORIENT " file)
              (display "\"" file)
              (display orient file)
              (display "\" )" file)
              (newline file)
              (display "      ( FIXED )" file)
              (newline file)
              (display "    )" file)
              (newline file)
              #t
            ) ; end begin
          ) ; end db-foreach
        )) ; end begin and if
        #t
      ) ; end begin
    ) ; end db-foreach
  ) ; end let*
)) ; end lambda and define

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpFixedCellPlace
;  Function: Dump the fixed standard cell location into a file in PDEF3.0
;  Last Modifed: 9/6/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpFixedCellPlace (lambda (cv file)
  (let*
    (
      (flag        0)
      (bBox      '())
      (transform  "")
      (orient    "0")
      (gateName   "")
      (x0          0)
      (y0          0)
      (info      '())
    )
    (db-foreach cv '() "cellInstMaster" master
      (begin
        (set! flag (dbFetchObjectField cv master "cellFlag"))
        (set! gateName (dbFetchObjectField cv master "cellName"))
        (if (> (bit-and flag dbbStandardCell) 0)
        (begin
          (db-foreach cv master "CellInst" cell
            (if (= (dbFetchObjectField cv cell "flag1") 2)
              (begin
                (set! info (getCellInstLocOrient cv cell master))
                (set! x0 (car info))
                (set! y0 (cadr info))
                (set! orient (caddr info))

                (display "    ( CELL " file)
                (display (dbFetchObjectField cv cell "name") file)
                (newline file)
                (display "      ( GATE_NAME " file)
                (display gateName file)
                (display " )" file)
                (newline file)
                (display "      ( LOC " file)
                (display (* dbResolution x0) file)
                (display " " file)
                (display (* dbResolution y0) file)
                (display " )" file)
                (newline file)
                (display "      ( ORIENT " file)
                (display "\"" file)
                (display orient file)
                (display "\" )" file)
                (newline file)
                (display "      ( RESTRICTION \"FIXED_PLACEMENT\" )" file)
                (newline file)
                (display "      ( FIXED )" file)
                (newline file)
                (display "    )" file)
                (newline file)
                #t
              ) ; end begin
              #t
            ) ; end if
          ) ; end db-foreach
        )) ; end begin and if
        #t
      ) ; end begin
    ) ; end db-foreach
  ) ; end let*
)) ; end lambda and define

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpCellPlace
;  Function: Dump the standard cell location into a file in PDEF3.0
;  Last Modifed: 9/6/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpCellPlace (lambda (cv file)
  (let*
    (
      (flag         0)
      (boundary   '())
      (transform   "")
      (orient      "")
      (bBox       '())
      (gateName    "")
      (x0           0)
      (y0           0)
      (info       '())
    )
    (db-foreach cv '() "cellInstMaster" master
      (begin
        (set! flag (dbFetchObjectField cv master "cellFlag"))
        (set! gateName (dbFetchObjectField cv master "cellName"))
        (if (> (bit-and flag dbbStandardCell) 0)
        (begin
          (db-foreach cv master "CellInst" cell
            (begin
              (set! info (getCellInstLocOrient cv cell master))
              (set! x0 (car info))
              (set! y0 (cadr info))
              (set! orient (caddr info))

              (display "    ( CELL " file)
              (display (dbFetchObjectField cv cell "name") file)
              (newline file)
              (display "      ( GATE_NAME " file)
              (display gateName file)
              (display " )" file)
              (newline file)
              (display "      ( LOC " file)
              (display (* dbResolution x0) file)
              (display " " file)
              (display (* dbResolution y0) file)
              (display " )" file)
              (newline file)
              (display "      ( ORIENT " file)
              (display "\"" file)
              (display orient file)
              (display "\" )" file)
              (newline file)
              (if (= (dbFetchObjectField cv cell "flag1") 2)
                (begin
                  (display "      ( RESTRICTION \"FIXED_PLACEMENT\" )" file)
                  (newline file)
                  (display "      ( FIXED )" file)
                  (newline file)
                )
              )
              (display "    )" file)
              (newline file)
              #t
            ) ; end begin
          ) ; end db-foreach
        )) ; end begin and if
        #t
      ) ; end begin
    ) ; end db-foreach
  ) ; end let*
)) ; end lambda and define


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpRegion
;  Function: Dump region and group information in PDEF3.0
;  Last Modifed: 6/28/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpRegion (lambda (cv file)
  (let*
    (
      (lib  (dbGetLibIdByCellId cv))
      (name           "")
      (region          0)
      (bbox          '())
      (master          0)
      (gateName       "")
    )
    (db-foreach cv '() "Group" grp
      (begin
        (set! name (dbFetchObjectField cv grp "name"))
        (set! region (dbFetchObjectField cv grp "parentId"))
        (if (not (null? region))
          (set! bbox (dbFetchObjectField cv region "bbox"))
          (set! bbox '())
        )
        (if (not (null? bbox))
          (begin
            (display "    ( CLUSTER " file)
            (display name file)
            (newline file)
            (display "      ( RECT " file)
            (display (* dbResolution (caar bbox)) file)
            (display " " file)
            (display (* dbResolution (cadar bbox)) file)
            (display " " file)
            (display (* dbResolution (caadr bbox)) file)
            (display " " file)
            (display (* dbResolution (cadadr bbox)) file)
            (display " )" file)
            (newline file)
            (display "      ( X_BOUNDS " file)
            (display (* dbResolution (caar bbox)) file)
            (display " " file)
            (display (* dbResolution (caadr bbox)) file)
            (display " )" file)
            (newline file)
            (display "      ( Y_BOUNDS " file)
            (display (* dbResolution (cadar bbox)) file)
            (display " " file)
            (display (* dbResolution (cadadr bbox)) file)
            (display " )" file)
            (newline file)

            (db-foreach cv grp "GroupMember" cell
              (begin
                (set! master (dbFetchObjectField cv cell "cellInstMasterId"))
                (set! gateName (dbFetchObjectField cv master "cellName"))
                (display "      ( CELL " file)
                (display (dbFetchObjectField cv cell "name") file)
                (newline file)
                (display "        ( GATE_NAME " file)
                (display gateName file)
                (display " )" file)
                (newline file)
                (display "      )" file)
                (newline file)
                #t
              ) ; end begin
            ) ; end db-foreach
            (display "    )" file)
            (newline file)
          ) ; end begin
        ) ; end if
      #t
      ) ; end begin
    ) ; end db-foreach
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpPlaceBlockage
;  Function: Dump the macro locations into a file in PDEF 3.0 format
;  Last Modifed: 5/26/99
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpPlaceBlockage (lambda (cv file)
  (let*
    (
      (blockageLayer  221)    ; Avant! placement blockage layer
      (bbox           '())
      (index            0)
    )
    (db-foreach cv '() "Rectangle" rect
      (begin
        (if (= (dbFetchObjectField cv rect "layer") blockageLayer)
          (begin
            (set! index (+ index 1))
            (set! bbox (dbFetchObjectField cv rect "bbox"))
            (display "    ( CLUSTER PLACEBLOCKAGE_" file)
            (display index file)
            (newline file)
            (display "      ( OBSTRUCTION " file)
            (display blockageLayer file)
            (display " )" file)
            (newline file)
            (display "      ( RECT " file)
            (display (* dbResolution (caar bbox)) file)
            (display " " file)
            (display (* dbResolution (cadar bbox)) file)
            (display " " file)
            (display (* dbResolution (caadr bbox)) file)
            (display " " file)
            (display (* dbResolution (cadadr bbox)) file)
            (display " )" file)
            (newline file)
            (display "      ( X_BOUNDS " file)
            (display (* dbResolution (caar bbox)) file)
            (display " " file)
            (display (* dbResolution (caadr bbox)) file)
            (display " )" file)
            (newline file)
            (display "      ( Y_BOUNDS " file)
            (display (* dbResolution (cadar bbox)) file)
            (display " " file)
            (display (* dbResolution (cadadr bbox)) file)
            (display " )" file)
            (newline file)
            (display "    )" file)
            (newline file)
          )
        )
        #t
      )
    )
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpPreRouteObs
;  Function: Dump the preroutes into a file in PDEF 3.0 format
;            Pre-routes are treated as placement or routing
;            obstructions, depending on the format flag
;  Last Modifed: 1/11/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpPreRouteObs (lambda (cv file format)
  (let*
    (
      (lib  (dbGetLibIdByCellId cv))
      (blockageLayer 221)    ; Avant! placement blockage layer
      (layer           0)
      (index           0)
      (blockageIndex   0)
      (len             0)
      (width           0)
      (ll            '())
      (ur            '())
    )
    (db-foreach cv '() "HorizontalWire" wire
      (begin
        (set! layer (dbFetchObjectField cv wire "layer"))
        (set! ll (car (dbFetchObjectField cv wire "startPoint")))
        (set! index (dbFetchObjectField cv wire "wireMasterIdx"))
        (set! width (vector-ref wireMasterList index))
        (set! len (dbFetchObjectField cv wire "length")) ; length in Avanti db unit
        (set! len (/ len dbResolution)) ; length in user unit (micron)
        (set! ll (list (car ll) (- (cadr ll) (/ width 2))))
        (set! ur (list (+ (car ll) len) (+ (cadr ll) width)))
        (cond
          ((= format 0) ; Treat pre routes as placement obstruction
            (begin
              (set! blockageIndex (+ blockageIndex 1))
              (display "    ( CLUSTER HORIZONTAL_WIRE_" file)
              (display blockageIndex file)
              (newline file)
              (display "      ( OBSTRUCTION " file)
              (display blockageLayer file)
              (display " )" file)
              (newline file)
              (display "      ( RECT " file)
              (display (* dbResolution (car ll)) file)
              (display " " file)
              (display (* dbResolution (cadr ll)) file)
              (display " " file)
              (display (* dbResolution (car ur)) file)
              (display " " file)
              (display (* dbResolution (cadr ur)) file)
              (display " )" file)
              (newline file)
              (display "      ( X_BOUNDS " file)
              (display (* dbResolution (car ll)) file)
              (display " " file)
              (display (* dbResolution (car ur)) file)
              (display " )" file)
              (newline file)
              (display "      ( Y_BOUNDS " file)
              (display (* dbResolution (cadr ll)) file)
              (display " " file)
              (display (* dbResolution (cadr ur)) file)
              (display " )" file)
              (newline file)
              (display "    )" file)
              (newline file)
            )
          )
          ((= format 1) ; treat pre routes as routing obstruction
            (begin
              (display "    ( OBSTRUCTION " file)
              (display layer file)
              (display " " file)
              (display (* dbResolution (car ll)) file)
              (display " " file)
              (display (* dbResolution (cadr ll)) file)
              (display " " file)
              (display (* dbResolution (car ur)) file)
              (display " " file)
              (display (* dbResolution (cadr ur)) file)
              (display " )" file)
              (newline file)
            )
          )
        )
        #t
      )
    )
    (set! blockageIndex 0)
    (db-foreach cv '() "VerticalWire" wire
      (begin
        (set! layer (dbFetchObjectField cv wire "layer"))
        (set! ll (car (dbFetchObjectField cv wire "startPoint")))
        (set! index (dbFetchObjectField cv wire "wireMasterIdx"))
        (set! width (vector-ref wireMasterList index))
        (set! len (dbFetchObjectField cv wire "length")) ; length in Avanti db unit
        (set! len (/ len dbResolution)) ; length in user unit (micron)
        (set! ll (list (- (car ll) (/ width 2)) (cadr ll)))
        (set! ur (list (+ (car ll) width) (+ (cadr ll) len)))
        (cond
          ((= format 0) ; treat pre routes as placement obstruction
            (begin
              (set! blockageIndex (+ blockageIndex 1))
              (display "    ( CLUSTER VERTICAL_WIRE_" file)
              (display blockageIndex file)
              (newline file)
              (display "      ( OBSTRUCTION " file)
              (display blockageLayer file)
              (display " )" file)
              (newline file)
              (display "      ( RECT " file)
              (display (* dbResolution (car ll)) file)
              (display " " file)
              (display (* dbResolution (cadr ll)) file)
              (display " " file)
              (display (* dbResolution (car ur)) file)
              (display " " file)
              (display (* dbResolution (cadr ur)) file)
              (display " )" file)
              (newline file)
              (display "      ( X_BOUNDS " file)
              (display (* dbResolution (car ll)) file)
              (display " " file)
              (display (* dbResolution (car ur)) file)
              (display " )" file)
              (newline file)
              (display "      ( Y_BOUNDS " file)
              (display (* dbResolution (cadr ll)) file)
              (display " " file)
              (display (* dbResolution (cadr ur)) file)
              (display " )" file)
              (newline file)
              (display "    )" file)
              (newline file)
            )
          )
          ((= format 1) ; treat pre routes as routing obstruction
            (begin
              (display "    ( OBSTRUCTION " file)
              (display layer file)
              (display " " file)
              (display (* dbResolution (car ll)) file)
              (display " " file)
              (display (* dbResolution (cadr ll)) file)
              (display " " file)
              (display (* dbResolution (car ur)) file)
              (display " " file)
              (display (* dbResolution (cadr ur)) file)
              (display " )" file)
              (newline file)
            )
          )
        )
        #t
      )
    )
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpPNET
;  Function: Dump routing wires into PNET section in IEEE PDEF
;  Last Modifed: 9/14/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpPNET (lambda (cv file)
  (let*
    (
      (lib  (dbGetLibIdByCellId cv))
      (layer           0)
      (index           0)
      (len             0)
      (width           0)
      (start         '())
      (end           '())
      (type            0)
      (contactCodeIdx  0)
      (xDuplicate      0)
      (yDuplicate      0)
      (xPitch          0)
      (yPitch          0)
      (center        '())
      (ret             0)
      (contactName    "")
      (master        '())
      (ul              0)
      (dir             0)
      (num             0)
      (netName        "")
    )
    (db-foreach cv '() "Net" net
      (begin
        (set! num 0)
        (if (isPowerGroundClockNet cv net)
          (begin
            (db-foreach cv net "NetMember" m
              (begin
                (set! ret (checkNetMember cv m))
                (if (and (= num 0) (> ret 0))
                  (begin
                    (set! pnetList (cons net pnetList))
                    (set! netName (dbFetchObjectField cv net "name"))
                    (display "    ( PNET " file)
                    (display netName file)
                    (newline file)
                    (set! type (dbFetchObjectField cv net "type"))
                    (cond
                      ((> (bit-and type dbbPowerNet) 0)
                        (display "      ( TYPE pwr )" file)
                      )
                      ((> (bit-and type dbbGroundNet) 0)
                        (display "      ( TYPE gnd )" file)
                      )
                      ((> (bit-and type dbbClockNet) 0)
                        (display "      ( TYPE clock )" file)
                      )
                      ((> (bit-and type dbbSplitClockNet) 0)
                        (display "      ( TYPE clock )" file)
                      )
                      ((> (bit-and type dbbSignalNet) 0)
                        (display "      ( TYPE data )" file)
                      )
                      (else
                        (display "      ( TYPE reserved )" file)
                      )
                    ) ; end cond
                    (newline file)
                  ) ; end begin
                ) ; end if
                (cond
                  ((= 1 ret) ; metal wire
                    (begin
                      (set! layer (dbFetchObjectField cv m "layer"))
                      (set! start (car (dbFetchObjectField cv m "startPoint")))
                      (set! index (dbFetchObjectField cv m "wireMasterIdx"))
                      (set! width (vector-ref wireMasterList index))
                      (set! len (dbFetchObjectField cv m "length")) ; in db unit
                      (set! len (/ len dbResolution)) ; in micron

                      (set! dir (getWireDirection cv m))
                      (cond
                        ((= dir 1) ; horizontal wire
                          (set! end (list (+ (car start) len) "*"))
                        )
                        ((= dir 2) ; vertical wire
                          (set! end (list "*" (+ (cadr start) len)))
                        )
                        (else
                          (display "ERROR: Fail to get direction of wire segment ")
                          (display m)
                          (display " of net ")
                          (display netName)
                          (display ". Wire segment is ignored.")
                          (newline)
                        )
                      )
                      (if (> dir 0)
                        (begin
                          (display "      ( ROUTE" file)
                          (newline file)
                          (display "        ( LAYER_WIDTH " file)
                          (display layer file)
                          (display " " file)
                          (display (* dbResolution width) file)
                          (display " )" file)
                          (newline file)
                          (display "        ( SHAPE  STRIPE )" file)
                          (newline file)
                          (display "        ( " file)
                          (display layer file)
                          (display " ( " file)
                          (display (* dbResolution (car start)) file)
                          (display " " file)
                          (display (* dbResolution (cadr start)) file)
                          (display " ) ( " file)
                          (if (equal? "*" (car end))
                            (display (car end) file)
                            (display (* dbResolution (car end)) file)
                          )
                          (display " " file)
                          (if (equal? "*" (cadr end))
                            (display (cadr end) file)
                            (display (* dbResolution (cadr end)) file)
                          )
                          (display " ) )" file)
                          (newline file)
                          (display "      )" file)
                          (newline file)
                        ) ; end begin
                      ) ; end if
                    ) ;end begin
                  )
                  ((= 2 ret) ; via
                    (begin
                      (set! contactCodeIdx (dbFetchObjectField cv m "contactCodeIdx"))
                      (set! contactName (getContactNameByContactCodeIdx lib contactCodeIdx))
                      (set! center (dbFetchObjectField cv m "centerPosition"))
                      (set! ul (dbFetchNonDefaultContactCodeInfo lib contactName "upperLayerNumber"))
                      (set! index 0)
                      (call-with-current-continuation (lambda (ret)
                        (for-each (lambda (v)
                          (begin
                            (set! index (+ index 1))
                            (if (= (car v) contactCodeIdx)
                              (ret index)
                            )
                          ))
                          viaMasterList
                        ) ;end for-each
                      )) ; end lambda and call

                      (display "      ( ROUTE" file)
                      (newline file)
                      (display "        ( LAYER_WIDTH " file)
                      (display ul file)
                      (display " 0.000 )" file)
                      (newline file)
                      (display "        ( " file)
                      (display ul file)
                      (display " ( " file)
                      (display (* dbResolution (caar center)) file)
                      (display " " file)
                      (display (* dbResolution (cadar center)) file)
                      (display " ) ( * * ) )" file)
                      (newline file)
                      (display "        ( VIA " file)
                      (display index file)
                      (display " ( " file)
                      (display (* dbResolution (caar center)) file)
                      (display " " file)
                      (display (* dbResolution (cadar center)) file)
                      (display " ) )" file)
                      (newline file)
                      (display "      )" file)
                      (newline file)
                    ) ;end begin
                  )
                  ((= 3 ret) ; via array
                    (begin
                      (set! contactCodeIdx (dbFetchObjectField cv m "contactCodeIdx"))
                      (set! contactName (getContactNameByContactCodeIdx lib contactCodeIdx))
                      (set! ul (dbFetchNonDefaultContactCodeInfo lib contactName "upperLayerNumber"))
                      (set! xDuplicate (dbFetchObjectField cv m "xDuplicate"))
                      (set! yDuplicate (dbFetchObjectField cv m "yDuplicate"))
                      (set! xPitch (dbFetchObjectField cv m "xPitch"))
                      (set! yPitch (dbFetchObjectField cv m "yPitch"))
                      (set! center (dbFetchObjectField cv m "position"))
                      (set! master (getViaArrayMaster cv m))
                      (if (null? master)
                        (begin
                          (display "WARNING: Cannot find any master for via array ")
                          (display m)
                          (newline)
                        )
                        (begin
                          (display "      ( ROUTE" file)
                          (newline file)
                          (display "        ( LAYER_WIDTH " file)
                          (display ul file)
                          (display " 0.000 )" file)
                          (newline file)
                          (display "        ( " file)
                          (display ul file)
                          (display " ( " file)
                          (display (* dbResolution (caar center)) file)
                          (display " " file)
                          (display (* dbResolution (cadar center)) file)
                          (display " ) ( * * ) )" file)
                          (newline file)
                          (display "        ( VIA " file)
                          (display (list-ref master 6) file)
                          (display " ( " file)
                          (display (* dbResolution (caar center)) file)
                          (display " " file)
                          (display (* dbResolution (cadar center)) file)
                          (display " ) )" file)
                          (newline file)
                          (display "      )" file)
                          (newline file)
                        )
                      ) ;end if
                    ) ;end begin
                  ) ; end via array
                ) ; end cond
                (if (> ret 0)
                  (set! num (+ num 1))
                )
                #t
              ) ; end begin
            ) ; end db-foreach
            (if (> num 0)
              (begin
                (display "    )" file)
                (newline file)
              )
            )
          ) ;end begin
          (set! netList (cons net netList)) ; signal net
        ) ; end if
        #t
      ) ; end begin
    ) ;end db-foreach
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpRoute
;  Function: Dump detail route of signal nets into NET
;            section in IEEE PDEF
;  Last Modifed: 9/13/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpRoute (lambda (cv file)
  (let*
    (
      (lib  (dbGetLibIdByCellId cv))
      (layer           0)
      (index           0)
      (len             0)
      (width           0)
      (start         '())
      (end           '())
      (netName        "")
      (contactCodeIdx  0)
      (xDuplicate      0)
      (yDuplicate      0)
      (xPitch          0)
      (yPitch          0)
      (center        '())
      (ret             0)
      (contactName    "")
      (master        '())
      (ul              0)
      (dir             0)
      (num             0)
      (netCount        0)
      (netIndex        0)
    )
    (set! netCount (length netList))

    (for-each (lambda (net)
      (begin ; signal nets
        (set! netIndex (+ netIndex 1))
        (cond
          ((= netIndex (round (* 0.1 netCount)))
            (display "    ...... 10%")(newline)
          )
          ((= netIndex (round (* 0.2 netCount)))
            (display "    ...... 20%")(newline)
          )
          ((= netIndex (round (* 0.3 netCount)))
            (display "    ...... 30%")(newline)
          )
          ((= netIndex netCount)
            (display "    ...... 100%")(newline)
          )
           ((= netIndex (round (* 0.4 netCount)))
            (display "    ...... 40%")(newline)
          )
          ((= netIndex (round (* 0.5 netCount)))
            (display "    ...... 50%")(newline)
          )
          ((= netIndex (round (* 0.6 netCount)))
            (display "    ...... 60%")(newline)
          )
          ((= netIndex (round (* 0.7 netCount)))
            (display "    ...... 70%")(newline)
          )
          ((= netIndex (round (* 0.8 netCount)))
            (display "    ...... 80%")(newline)
          )
          ((= netIndex (round (* 0.9 netCount)))
            (display "    ...... 90%")(newline)
          )
          ((= netIndex netCount)
            (display "    ...... 100%")(newline)
          )
       )
        (set! netName (dbFetchObjectField cv net "name"))
        (set! num 0)
        (db-foreach cv net "NetMember" m
          (begin
            (set! ret (checkNetMember cv m))
            (if (and (= 0 num) (> ret 0))
              (begin
                (display "    ( NET " file)
                (display netName file)
                (newline file)
              )
            )
            (cond
              ((= 1 ret) ; metal wire
                (begin
                  (set! layer (dbFetchObjectField cv m "layer"))
                  (set! start (car (dbFetchObjectField cv m "startPoint")))
                  (set! index (dbFetchObjectField cv m "wireMasterIdx"))
                  (set! width (vector-ref wireMasterList index))
                  (set! len (dbFetchObjectField cv m "length")) ; in db unit
                  (set! len (/ len dbResolution)) ; in micron

                  (set! dir (getWireDirection cv m))
                  (cond
                    ((= dir 1) ; horizontal wire
                      (set! end (list (+ (car start) len) "*"))
                    )
                    ((= dir 2) ; vertical wire
                      (set! end (list "*" (+ (cadr start) len)))
                    )
                    (else
                      (display "ERROR: Fail to get direction of wire segment ")
                      (display m)
                      (display " of net ")
                      (display netName)
                      (display ". Wire segment is ignored.")
                      (newline)
                    )
                  )
                  (if (> dir 0)
                    (begin
                      (display "      ( ROUTE" file)
                      (newline file)
                      (display "        ( LAYER_WIDTH " file)
                      (display layer file)
                      (display " " file)
                      (display (* dbResolution width) file)
                      (display " )" file)
                      (newline file)
                      (display "        ( SHAPE  STRIPE )" file)
                      (newline file)
                      (display "        ( " file)
                      (display layer file)
                      (display " ( " file)
                      (display (* dbResolution (car start)) file)
                      (display " " file)
                      (display (* dbResolution (cadr start)) file)
                      (display " ) ( " file)
                      (if (equal? "*" (car end))
                        (display (car end) file)
                        (display (* dbResolution (car end)) file)
                      )
                      (display " " file)
                      (if (equal? "*" (cadr end))
                        (display (cadr end) file)
                        (display (* dbResolution (cadr end)) file)
                      )
                      (display " ) )" file)
                      (newline file)
                      (display "      )" file)
                      (newline file)
                    ) ; end begin
                  ) ; end if
                ) ;end begin
              )
              ((= 2 ret) ; via
                (begin
                  (set! contactCodeIdx (dbFetchObjectField cv m "contactCodeIdx"))
                  (set! contactName (getContactNameByContactCodeIdx lib contactCodeIdx))
                  (set! center (dbFetchObjectField cv m "centerPosition"))
                  (set! ul (dbFetchNonDefaultContactCodeInfo lib contactName "upperLayerNumber"))
                  (set! index 0)
                  (call-with-current-continuation (lambda (ret)
                    (for-each (lambda (v)
                      (begin
                        (set! index (+ index 1))
                        (if (= (car v) contactCodeIdx)
                          (ret index)
                        )
                      ))
                      viaMasterList
                    ) ;end for-each
                  )) ; end lambda and call

                  (display "      ( ROUTE" file)
                  (newline file)
                  (display "        ( LAYER_WIDTH " file)
                  (display ul file)
                  (display " 0.000 )" file)
                  (newline file)
                  (display "        ( " file)
                  (display ul file)
                  (display " ( " file)
                  (display (* dbResolution (caar center)) file)
                  (display " " file)
                  (display (* dbResolution (cadar center)) file)
                  (display " ) ( * * ) )" file)
                  (newline file)
                  (display "        ( VIA " file)
                  (display index file)
                  (display " ( " file)
                  (display (* dbResolution (caar center)) file)
                  (display " " file)
                  (display (* dbResolution (cadar center)) file)
                  (display " ) )" file)
                  (newline file)
                  (display "      )" file)
                  (newline file)
                ) ;end begin
              )
              ((= 3 ret) ; via array
                (begin
                  (set! contactCodeIdx (dbFetchObjectField cv m "contactCodeIdx"))
                  (set! contactName (getContactNameByContactCodeIdx lib contactCodeIdx))
                  (set! ul (dbFetchNonDefaultContactCodeInfo lib contactName "upperLayerNumber"))
                  (set! xDuplicate (dbFetchObjectField cv m "xDuplicate"))
                  (set! yDuplicate (dbFetchObjectField cv m "yDuplicate"))
                  (set! xPitch (dbFetchObjectField cv m "xPitch"))
                  (set! yPitch (dbFetchObjectField cv m "yPitch"))
                  (set! center (dbFetchObjectField cv m "position"))
                  (set! master (getViaArrayMaster cv m))
                  (if (null? master)
                    (begin
                      (display "WARNING: Cannot find any master for via array ")
                      (display m)
                      (newline)
                    )
                    (begin
                      (display "      ( ROUTE" file)
                      (newline file)
                      (display "        ( LAYER_WIDTH " file)
                      (display ul file)
                      (display " 0.000 )" file)
                      (newline file)
                      (display "        ( " file)
                      (display ul file)
                      (display " ( " file)
                      (display (* dbResolution (caar center)) file)
                      (display " " file)
                      (display (* dbResolution (cadar center)) file)
                      (display " ) ( * * ) )" file)
                      (newline file)
                      (display "        ( VIA " file)
                      (display (list-ref master 6) file)
                      (display " ( " file)
                      (display (* dbResolution (caar center)) file)
                      (display " " file)
                      (display (* dbResolution (cadar center)) file)
                      (display " ) )" file)
                      (newline file)
                      (display "      )" file)
                      (newline file)
                    )
                  ) ;end if
                ) ;end begin
              ) ; end via array
            ) ; end cond
            (if (> ret 0)
              (set! num (+ num 1))
            )
            #t
          ) ; end begin
        ) ; end db-foreach
        (if (> num 0)
          (begin
            (display "    )" file)
            (newline file)
          )
        )
      )) ; end begin
      netList
    ) ; end for-each
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: getCellInstLocOrient
;  Function: Get the lower left corner coordiate as well as
;            orientation of the cell instance
;  Last Modifed: 9/6/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define getCellInstLocOrient (lambda (cv cell master)
  (let
    (
      (lib  (dbGetLibIdByCellId cv))
      (x0         0)
      (y0         0)
      (transform  0)
      (orient    "")
      (origin   '())
      (bbox     '())
      (llx        0)
      (lly        0)
      (urx        0)
      (ury        0)
      (width      0)
      (height     0)
      (dx         0)
      (dy         0)
    )
    (if (= 0 (dbFetchObjectField cv master "prBoundaryId"))
      (set! bbox (dbFetchObjectField cv master "cellBoundaryBBox"))
      (set! bbox (dbFetchObjectField cv master "prBoundaryBBox"))
    )
    (set! width (abs (- (caadr bbox) (caar bbox))))
    (set! height (abs (- (cadadr bbox) (cadar bbox))))
    (set! dx (- 0 (caar bbox))) ; delta x of lower left corner and origin
    (set! dy (- 0 (cadar bbox))) ; delta y of lower left corner and origin

    (set! origin (car (dbFetchObjectField cv cell "position")))
    ; position is the location of the origin of the cell
    ; origin is not necessarily the lower left corner of the cell
    (set! x0 (car origin))
    (set! y0 (cadr origin))

    (set! transform (dbFetchObjectField cv cell "transformCode"))
    (cond
      ((= transform dbcRotate0)
        (begin
          (set! orient "0")
          (set! x0 (- x0 dx))
          (set! y0 (- y0 dy))
        )
      )
      ((= transform dbcRotate90)
        (begin
          (set! orient "90")
          (set! x0 (- (+ x0 dy) height))
          (set! y0 (- y0 dx))
        )
      )
      ((= transform dbcRotate180)
        (begin
          (set! orient "180")
          (set! x0 (- (+ x0 dx) width))
          (set! y0 (- (+ y0 dy) height))
        )
      )
      ((= transform dbcRotate270)
        (begin
          (set! orient "270")
          (set! x0 (- x0 dy))
          (set! y0 (- (+ y0 dx) width))
        )
      )
      ((= transform dbcRotate0MirrorX)
        (begin
          (set! orient "0-mirror")
          (set! x0 (- (+ x0 dx) width))
          (set! y0 (- y0 dy))
        )
      )
      ((= transform dbcRotate90MirrorX)
        (begin
          (set! orient "90-mirror")
          (set! x0 (- x0 dy))
          (set! y0 (- y0 dx))
        )
      )
      ((= transform dbcRotate180MirrorX)
        (begin
          (set! orient "180-mirror")
          (set! x0 (- x0 dx))
          (set! y0 (- (+ y0 dy) height))
        )
      )
      ((= transform dbcRotate270MirrorX)
        (begin
          (set! orient "270-mirror")
          (set! x0 (- (+ x0 dy) height))
          (set! y0 (- (+ y0 dx) width))
        )
      )
      ((= transform dbcRotate0MirrorY)
        (begin
          (set! orient "180-mirror")
          (set! x0 (- x0 dx))
          (set! y0 (- (+ y0 dy) height))
        )
      )
      ((= transform dbcRotate90MirrorY)
        (begin
          (set! orient "270-mirror")
          (set! x0 (- (+ x0 dy) height))
          (set! y0 (- (+ y0 dx) width))
        )
      )
      ((= transform dbcRotate180MirrorY)
        (begin
          (set! orient "0-mirror")
          (set! x0 (- (+ x0 dx) width))
          (set! y0 (- y0 dy))
        )
      )
      ((= transform dbcRotate270MirrorY)
        (begin
          (set! orient "90-mirror")
          (set! x0 (- x0 dy))
          (set! y0 (- y0 dx))
        )
      )
    ) ; end cond
    (set! x0 (roundoff x0 dbResolution))
    (set! y0 (roundoff y0 dbResolution))
    (list x0 y0 orient)
  )
))

(define checkNetMember (lambda (cv m)
  (let*
    (
      (record '())
      (ret 0)
    )
    (if (dbId? m)
      (begin
        (set! record (dbFetchObject cv m))

        (cond
          ((equal? "wireMasterIdx" (caadr record))
            (set! ret 1) ; wire
          )
          ((equal? "contactCodeIdx" (caadr record))
            (set! ret 2) ; regular via
          )
          ((equal? "bitmap" (caar (reverse record)))
            (set! ret 3) ; via array
          )
          (else
            (set! ret 0)
          )
        ) ; end cond
      ) ; begin
    ) ; end if
    ret
  )
))

(define getWireDirection (lambda (cv wire)
  (let*
    (
      (dir 0)
    )
    (if (member wire horizontalWireList)
      (set! dir 1)
      (set! dir 2)
    )
    dir
  )
))


(define isPowerGroundClockNet (lambda (cv net)
  (or
    (or
      (> (bit-and (dbFetchObjectField cv net "type") dbbClockNet) 0)
      (> (bit-and (dbFetchObjectField cv net "type") dbbSplitClockNet) 0)
    )
    (or
      (> (bit-and (dbFetchObjectField cv net "type") dbbPowerNet) 0)
      (> (bit-and (dbFetchObjectField cv net "type") dbbGroundNet) 0)
    )
  )
))

(define dumpFP (lambda (filename)
  (axgDumpFloorPlan)
  (setFormField "Dump Floor Plan" "Output Filename" filename)
  (setToggleField "Dump Floor Plan" "Type" "io cell placement" 1)
  (setToggleField "Dump Floor Plan" "Type" "std cell placement" 0)
  (setToggleField "Dump Floor Plan" "Type" "macro cell placement" 1)
  (setToggleField "Dump Floor Plan" "Type" "row" 1)
  (setToggleField "Dump Floor Plan" "Type" "track" 1)
  (setToggleField "Dump Floor Plan" "Type" "region" 1)
  (setToggleField "Dump Floor Plan" "Type" "Floor Plan group" 1)
  (setToggleField "Dump Floor Plan" "Type" "Blockage/Density" 1)
  (setFormField "Dump Floor Plan" "No Backslashes in Cell Instance Names" "0")
  (formOK "Dump Floor Plan")
))


(define add-dumpFP-log (lambda (in_port out_port)
  (let f
    ((str (read-string in_port)))
    (cond
      ((equal? '() str) #t)
      ((equal? #\nl (peek-char in_port))
        (display str out_port)
        (newline out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "define" str)
        (display "NAME" out_port)
        (display " " out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "_cell" str)
        (display "DESIGN" out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "dbCreateCellBoundary" str)
        (display "BBOX" out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "dbRestorePin" str)
        (display "PIN" out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "dbSetCellInstPlacement" str)
        (display "MODULE" out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "axPurgeSingleRecordType" str)
        (display "REMOVE" out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "axCreateBaseArrayRecord" str)
        (display "COREAREA" out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "axCreateTilePatternRecord" str)
        (display "TILEPATTERN" out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "axCreateRowRecord" str)
        (display "CELLROW" out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "axCreateWireDirRecord" str)
        (display "WIREDIR" out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "axCreateTrackRecord" str)
        (display "TRACK" out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
      ((equal? "(dbCreateRectangle" str)
        (display "(RECTANGLE" out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
      (else
        (display str out_port)
        (display "      " out_port)
        (f (set! str (read-string in_port)))
      )
    )
  )
))

(define roundoff (lambda (number resolution)
  (/ (round (* number resolution)) resolution)
))

(define replace-string-char (lambda (str char newChar)
  (let
    (
      (newStr "")
      (len 0)
    )
    (set! newStr (string-copy str))
    (set! len (string-length str))
    (do
      ((i 0 (+ i 1)))
      ((equal? i len))
      (if (equal? char (string-ref str i))
        (string-set! newStr i newChar)
      )
    )
    newStr
  )
))

(define busNameRule (lambda (name)
  (let
    ((newName ""))
    (set! newName (string-copy name))
    (cond
      ((equal? pdefBusNameRule "[%d]")
        (set! newName (replace-string-char newName #\( #\[))
        (set! newName (replace-string-char newName #\) #\]))
        (set! newName (replace-string-char newName #\< #\[))
        (set! newName (replace-string-char newName #\> #\]))
        (set! newName (replace-string-char newName #\{ #\[))
        (set! newName (replace-string-char newName #\} #\]))
      )
      ((equal? pdefBusNameRule "<%d>")
        (set! newName (replace-string-char newName #\( #\<))
        (set! newName (replace-string-char newName #\) #\>))
        (set! newName (replace-string-char newName #\[ #\<))
        (set! newName (replace-string-char newName #\] #\>))
        (set! newName (replace-string-char newName #\{ #\<))
        (set! newName (replace-string-char newName #\} #\>))
      )
      ((equal? pdefBusNameRule "(%d)")
        (set! newName (replace-string-char newName #\[ #\())
        (set! newName (replace-string-char newName #\] #\)))
        (set! newName (replace-string-char newName #\< #\())
        (set! newName (replace-string-char newName #\> #\)))
        (set! newName (replace-string-char newName #\{ #\())
        (set! newName (replace-string-char newName #\} #\)))
      )
      ((equal? pdefBusNameRule "{%d}")
        (set! newName (replace-string-char newName #\( #\{))
        (set! newName (replace-string-char newName #\) #\}))
        (set! newName (replace-string-char newName #\< #\{))
        (set! newName (replace-string-char newName #\> #\}))
        (set! newName (replace-string-char newName #\[ #\{))
        (set! newName (replace-string-char newName #\] #\}))
      )
    )
    newName
  )
))

(define getViaArrayMaster (lambda (cv via)
  (let*
    (
      (idx             0)
      (col             0)
      (row             0)
      (x               0)
      (y               0)
      (match         '())
      (flag           #f)
      (masterList    '())
      (nth           '())
    )
    (set! idx (dbFetchObjectField cv via "contactCodeIdx"))
    (set! col (dbFetchObjectField cv via "xDuplicate"))
    (set! row (dbFetchObjectField cv via "yDuplicate"))
    (set! x (dbFetchObjectField cv via "xPitch"))
    (set! y (dbFetchObjectField cv via "yPitch"))

    (call-with-current-continuation (lambda (break)
      (do
        ((i 0 (+ i 1)))
        ((= i (length contactCodeIndexList)))
        (if (= (list-ref contactCodeIndexList i) idx)
          (begin
            (set! nth i)
            (break nth)
          )
        )
      )
    ))

    (if (not (null? nth))
      (begin
        (set! masterList (vector-ref viaArrayMasterList nth))
        (call-with-current-continuation (lambda (ret)
          (for-each (lambda (master)
            (begin
              (call-with-current-continuation (lambda (cont)
                (if (not (= (car master) idx))
                  (begin
                    (set! flag #f)
                    (cont flag)
                  )
                  (set! flag #t)
                )
                (if (not (= (cadr master) col))
                  (begin
                    (set! flag #f)
                    (cont flag)
                  )
                  (set! flag #t)
                )
                (if (not (= (list-ref master 2) row))
                  (begin
                    (set! flag #f)
                    (cont flag)
                  )
                  (set! flag #t)
                )
                (if (not (= (list-ref master 3) x))
                  (begin
                    (set! flag #f)
                    (cont flag)
                  )
                  (set! flag #t)
                )
                (if (not (= (list-ref master 4) y))
                  (begin
                    (set! flag #f)
                    (cont flag)
                  )
                  (set! flag #t)
                )
              )) ; end lambda and call

              (if (equal? #t flag)
                (begin
                  (set! match master)
                  (ret match)
                )
                (begin
                  (set! match '())
                  (set! flag #t)
                )
              ) ; end if
            )) ; end begin and lambda
            masterList
          ) ; end for-each
        )) ; end lambda and call
      ) ; end begin
    ) ; end if
    match
  )
))

(define string-parse (lambda (str)
  (let
    (
      (strList '())
      (numList '())
      (number  '())
      (len       0)
      (index     0)
    )
    (set! strList (string->list str))
    (set! len (length strList))
    (for-each (lambda (char)
      (cond
        ((equal? #\space char)
          (begin
            (if (not (null? number))
              (begin
                (if (equal? #f (string->number (list->string (reverse number))))
                  (set! number (list->string (reverse number)))
                  (set! number (string->number (list->string (reverse number))))
                )
                (set! numList (cons number numList))
                (set! number '())
              )
            )
            (set! index (+ index 1))
          )
        )
        ((= index (- len 1))
          (begin
            (set! number (cons char number))
            (if (equal? #f (string->number (list->string (reverse number))))
              (set! number (list->string (reverse number)))
              (set! number (string->number (list->string (reverse number))))
            )
            (set! numList (cons number numList))
          )
        )
        (else
          (begin
            (set! number (cons char number))
            (set! index (+ index 1))
          )
        )
      )) ; end cond and lambda
      strList
    ) ; end for-each
    (reverse numList)
  )
))

