
(define plibDumpMode 2) ; 0 - technology; 1 - cell; 2 - technology + cell
(define plibBusNameRule "[%d]")
(define plibSiteName "core")
(define nameCaseSensitivity "ON")
(define maxFatWireWidth 2500)
(define debugFlag #f)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpPLIB
;  Function: Dump Avanti library into PLIB
;  Last Modifed: 9/9/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpPLIB (lambda (libName fileName)
  (let
    (
      (lib 0)
      (file (open-output-file fileName))
    )
    ; reset global variables
    (define routeLayerNumList '())
    (define viaLayerNumList   '())

    (if (equal? #t debugFlag)
      (define dumpPLIBLogFile (open-output-file "dumpPLIB.log"))
    )
    (if (not (equal? #f (dbGetCurrentLibId)))
      (begin
        (display "Close previously opened library.")
        (newline)
        (geCloseLib)
      )
    )
    (display "Open library ")
    (display libName)
    (newline)
    (set! lib (dbOpenLibrary libName "r"))
    (if (equal? #f lib)
      (begin
        (display "ERROR: Cannot find library ")
        (display libName)
        (newline)
      )
      (begin
        (display "phys_library ( \"" file)
        (display libName file)
        (display "\" ) {" file)
        (newline file)
        (case plibDumpMode
          ((0)
            (dumpTech lib file)
          )
          ((1)
            (dumpLib lib file)
          )
          ((2)
            (begin
              (dumpTech lib file)
              (dumpLib lib file)
            )
          )
        )
        (display "} /* end phys_library */" file)
        (newline file)
        (close-output-port file)
        (if (equal? #t debugFlag)
          (begin
            (close-output-port dumpPLIBLogFile)
            (dbDumpTechFile lib ".tech.dump")
            (clfDump lib ".clf.dump")
            (system "/bin/cat dumpPLIB.log .tech.dump .clf.dump > tmp.log")
            (system "/bin/mv tmp.log dumpPLIB.log")
            (system "/bin/rm -f .clf.dump")
            (system "/bin/rm -f .tech.dump")
          )
        )
        (dbCloseLib lib)
      )
    )
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpTech
;  Function: Dump the Avanti technology information
;  Last Modifed: 9/7/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpTech (lambda (lib file)
  (let
    ((test 0))

    (display "Dump technology section")(newline)
    (display "    Unit definition...")(newline)
    (dumpUnit lib file)
    (display "    Layer definition...")(newline)
    (display "  resource ( \"std_cell\" ) {" file)
    (newline file)
    (dumpLayer lib file)
    (display "    Via definition...")(newline)
    (dumpVia lib file)
    (display "    Site definition...")(newline)
    (dumpSite lib file)
    (display "  } /* end resource */" file)
    (newline file)
    (newline file)
    (display "    Same-net spacing definition...")(newline)
    (display "  topological_design_rules () {" file)
    (newline file)
    (dumpSameNetSpacing lib file)
    (display "    Via rule generate definition...")(newline)
    (dumpViaGen lib file)
    (display "  } /* end toplogical_design_rules */" file)
    (newline file)
    (newline file)
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpUnit
;  Function: Dump unit section
;  Last Modifed: 9/7/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpUnit (lambda (lib file)
  (let
    (
      (timeUnit       "")
      (lengthUnit     "")
      (lengthPrecision 0)
      (resUnit        "")
      (capUnit        "")
      (powerUnit      "")
      (voltUnit       "")
      (currentUnit    "")
      (siteName       "")
      (layerName      "")
      (maskName       "")
      (width           0)
      (spacing         0)
      (pitch           0)
    )

    (set! timeUnit (dbFetchTechInfo lib "unitTimeName"))
    (set! lengthUnit (dbFetchTechInfo lib "unitLengthName"))
    (set! lengthPrecision (dbFetchTechInfo lib "lengthPrecision"))
    (set! resUnit (dbFetchTechInfo lib "unitResistanceName"))
    (set! capUnit (dbFetchTechInfo lib "unitCapacitanceName"))
    (set! powerUnit (dbFetchTechInfo lib "unitPowerName"))
    (set! voltUnit (dbFetchTechInfo lib "unitVoltageName"))
    (set! currentUnit (dbFetchTechInfo lib "unitCurrentName"))

    (if (not (equal? timeUnit ""))
      (begin
        (display "  time_unit : " file)
        (cond
          ;((equal? timeUnit "ms")
          ;  (display "1ms;" file)
          ;)
          ;((equal? timeUnit "us")
          ;  (display "1us;" file)
          ;)
          ((equal? timeUnit "ns")
            (display "1ns;" file)
          )
          ((equal? timeUnit "ps")
            (display "1ps;" file)
          )
          (else
            (begin
              (display "WARNING: Unrecognized time unit ")
              (display timeUnit file)
              (newline)
            )
          )
        )
        (newline file)
      )
    )

    (if (not (equal? capUnit ""))
      (begin
        (display "  capacitance_unit : " file)
        (cond
          ;((equal? capUnit "f")
          ;  (display "1f;" file)
          ;)
          ;((equal? capUnit "uf")
          ;  (display "1mf;" file)
          ;)
          ((equal? capUnit "pf")
            (display "1pf;" file)
          )
          ((equal? capUnit "ff")
            (display "1ff;" file)
          )
          (else
            (begin
              (display "WARNING: Unrecognized capcitance unit ")
              (display capUnit)
              (newline)
            )
          )
        )
        (newline file)
      )
    )

    (if (not (equal? resUnit ""))
      (begin
        (display "  resistance_unit : " file)
        (cond
          ((equal? resUnit "kohm")
            (display "1kohm;" file)
          )
          ((equal? resUnit "ohm")
            (display "1ohm;" file)
          )
          (else
            (begin
              (display "WARNING: Unrecognized resistance unit ")
              (display resUnit)
              (newline)
            )
          )
        )
        (newline file)
      )
    )

    (if (not (equal? powerUnit ""))
      (begin
        (display "  power_unit : " file)
        (cond
          ;((equal? powerUnit "w")
          ;  (display "1w;" file)
          ;)
          ((equal? powerUnit "mw")
            (display "1mW;" file)
          )
          ;((equal? powerUnit "uw")
          ;  (display "1uw;" file)
          ;)
          ;((equal? powerUnit "nw")
          ;  (display "1 nw;" file)
          ;)
          ;((equal? powerUnit "pw")
          ;  (display "1pw;" file)
          ;)
          (else
            (begin
              (display "WARNING: Unrecognized power unit ")
              (display powerUnit)
              (newline)
           )
          )
        )
        (newline file)
      )
    )

    (if (not (equal? currentUnit ""))
      (begin
        (display "  current_unit : " file)
        (cond
          ((or (equal? currentUnit "A") (equal? currentUnit "a"))
            (display "1A;" file)
          )
          ((or (equal? currentUnit "mA") (equal? currentUnit "ma"))
            (display "1mA;" file)
          )
          ((or (equal? currentUnit "uA") (equal? currentUnit "ua"))
            (display "1uA;" file)
          )
          ;((or (equal? currentUnit "nA") (equal? currentUnit "na"))
          ;  (display "1nA;" file)
          ;)
          ;((or (equal? currentUnit "pA") (equal? currentUnit "pa"))
          ;  (display "1pA;" file)
          ;)
          (else
            (begin
              (display "WARNING: Unrecognized current unit ")
              (display currentUnit)
              (newline)
            )
          )
        )
        (newline file)
      )
    )

    (if (not (equal? voltUnit ""))
      (begin
        (display "  voltage_unit : " file)
        (cond
          ((or (equal? voltUnit "V") (equal? voltUnit "v"))
            (display "1V;" file)
          )
          ((or (equal? voltUnit "mV") (equal? voltUnit "mv"))
            (display "1mv;" file)
          )
          ;((or (equal? voltUnit "uV") (equal? voltUnit "uv"))
          ;  (display "1uV;" file)
          ;)
          ;((or (equal? voltUnit "nV") (equal? voltUnit "nv"))
          ;  (display "1nV;" file)
          ;)
          ;((or (equal? voltUnit "pV") (equal? voltUnit "pv"))
          ;  (display "1pV;" file)
          ;)
          (else
            (begin
              (display "WARNING: Unrecognized voltage unit ")
              (display voltUnit)
              (newline)
            )
          )
        )
        (newline file)
      )
    )

    (if (not (equal? lengthUnit ""))
      (begin
        (display "  distance_unit : " file)
        (cond
          ((equal? lengthUnit "micron")
            (display "1um;" file)
          )
          ((equal? lengthUnit "mil")
            (display "1mm;" file)
          )
        )
      )
    )
    (newline file)

    (display "  dist_conversion_factor : " file)
    (display lengthPrecision file)
    (display ";" file)
    (newline file)

    (display "  bus_naming_style : " file)
    (cond
      ((equal? "[%d]" plibBusNameRule)
        (display "\"%s[%d]\";" file)
      )
      ((equal? "<%d>" plibBusNameRule)
        (display "\"%s<%d>\";" file)
      )
      ((equal? "(%d)" plibBusNameRule)
        (display "\"%s(%d)\";" file)
      )
      ((equal? "{%d}" plibBusNameRule)
        (display "\"%s{%d}\";" file)
      )
    )
    (newline file)
    (newline file)
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpLayer
;  Function: Dump layer section
;  Last Modifed: 10/26/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpLayer (lambda (lib file)
  (let
    (
      (layerList (dbFetchLayerIdList lib))
      (maskList (list "poly" "polyCont" "metal1" "via1" "metal2"
        "via2" "metal3" "via3" "metal4" "via4" "metal5" "via5" "metal6"
        "via6" "metal7" "via7" "metal8" "via8" "metal9" "via9" "metal10"))
      (routeList (list "poly" "metal1" "metal2" "metal3" "metal4" "metal5"
        "metal6" "metal7" "metal8" "metal9" "metal10"))
      (contList  (list "polyCont" "via1" "via2" "via3" "via4" "via5" "via6"
        "via7" "via8" "via9"))
      (hList '("metal1" "metal3" "metal5" "metal7" "metal9"))
      (vList '("metal2" "metal4" "metal6" "metal8" "metal10"))
      (layerNumList      '())
      (capPrecision        0)
      (resPrecision        0)
      (layerNum            0)
      (layerName          "")
      (maskName           "")
      (width               0)
      (pitch               0)
      (spacing             0)
      (fatThreshold        0)
      (fatSpacing          0)
      (dir                "")
      (areaCap             0)
      (sideWallCap         0)
      (edgeCap             0)
      (intraCap            0)
      (interCap            0)
      (res                 0)
    )

    (set! capPrecision (dbFetchTechInfo lib "capacitancePrecision"))
    (set! resPrecision (dbFetchTechInfo lib "resistancePrecision"))

    (for-each (lambda (layer)
      (begin
        (set! maskName (dbFetchLayerInfo lib layer "maskName"))
        (if (not (equal? maskName ""))
          (cond
            ((member maskName routeList)
              (set! routeLayerNumList (cons layer routeLayerNumList))
            )
            ((member maskName contList)
              (set! viaLayerNumList (cons layer viaLayerNumList))
            )
          )
        )
      ))
      layerList
    )
    (set! routeLayerNumList (reverse routeLayerNumList))
    (set! viaLayerNumList (reverse viaLayerNumList))
    (set! layerList (append routeLayerNumList viaLayerNumList))

    (for-each (lambda (mask)
      (begin
        (for-each (lambda (layer)
          (begin
            (set! maskName (dbFetchLayerInfo lib layer "maskName"))
            (if (equal? maskName mask)
              (set! layerNumList (cons layer layerNumList))
            )
          ))
          layerList
        )
      ))
      maskList
    )
    (set! layerNumList (reverse layerNumList))

    (for-each (lambda (layerNum)
      (begin
        (if (member layerNum routeLayerNumList)
          (begin ; metal
            (set! layerName (dbGetLayerNameFromNumber lib layerNum))
            (set! width (dbFetchLayerInfo lib layerNum "defaultWidth"))
            (set! spacing (dbFetchLayerInfo lib layerNum "minSpacing"))
            (set! fatThreshold (dbFetchLayerInfo lib layerNum "fatWireThreshold"))
            (set! fatSpacing (dbFetchLayerInfo lib layerNum "fatFatMinSpacing"))
            (set! pitch (dbFetchLayerInfo lib layerNum "pitch"))
            (set! maskName (dbFetchLayerInfo lib layerNum "maskName"))
            (set! dir (dbFetchLayerInfo lib layerNum "preferredDirection"))
            (if (or (= dir 0) (null? dir))
              (cond
                ((member maskName hList)
                  (set! dir "horizontal")
                )
                ((member maskName vList)
                  (set! dir "vertical")
                )
              )
            )
            (set! res (dbFetchLayerInfo lib layerNum "unitMaxResistance"))
            (set! areaCap (dbFetchLayerInfo lib layerNum "unitMaxCapacitance"))
            (set! sideWallCap (dbFetchLayerInfo lib layerNum "unitMaxSideWallCap"))
            (set! intraCap (dbFetchFringeCapInfo lib layerNum layerNum "maxFringeCap"))
            (if (not (null? intraCap))
              (begin
                (set! intraCap (/ (* 0.4 intraCap) capPrecision))
                (set! edgeCap (+ (/ sideWallCap 2) intraCap))
              )
              (set! edgeCap (/ sideWallCap 2))
            )
            ; assume average adjacency ratio 50% for intra-layer fringe
            ; intra-layer fringe is considered as part of side wall cap

            (for-each (lambda (other_layer)
              (if (not (= layerNum other_layer))
                (begin
                  (set! interCap (dbFetchFringeCapInfo lib layerNum other_layer "maxFringeCap"))
                  (if (not (null? interCap))
                    (begin
                      (set! interCap (* 0.3 (/ interCap capPrecision)))
                      (set! areaCap (+ areaCap interCap))
                    )
                  )
                )
              ))
              routeLayerNumList
            )
            (if (not (equal? maskName "poly"))
              (begin
                (display "    routing_layer ( \"" file)
                (display layerName file)
                (display "\" ) {" file)
                (newline file)
                (display "      default_routing_width : " file)
                (display width file)
                (display ";" file)
                (newline file)
                (if (> fatThreshold 0)
                  (begin
                    (display "      ranged_spacing ( " file)
                    (display "0.0, " file)
                    (display fatThreshold file)
                    (display ", " file)
                    (display spacing file)
                    (display " );" file)
                    (newline file)
                    (display "      ranged_spacing ( " file)
                    (display fatThreshold file)
                    (display ", " file)
                    (display maxFatWireWidth file)
                    (display ", " file)
                    (display fatSpacing file)
                    (display " );" file)
                  )
                  (begin
                    (display "      spacing : " file)
                    (display spacing file)
                    (display ";" file)
                  )
                )
                (newline file)
                (display "      routing_direction : " file)
                (display dir file)
                (display ";" file)
                (newline file)
                (display "      pitch : " file)
                (display pitch file)
                (display ";" file)
                (newline file)
                (display "      res_per_sq : " file)
                (display res file)
                (display ";" file)
                (newline file)
                (display "      cap_per_sq : " file)
                (display areaCap file)
                (display ";" file)
                (newline file)
                (display "      edgecapacitance : " file)
                (display edgeCap file)
                (display ";" file)
                (newline file)
                (display "    } /* end routing_layer */" file)
                (newline file)
              )
              (begin
                (display "    device_layer ( \"" file)
                (display layerName file)
                (display "\" ) ;" file)
                (newline file)
              )
            ) ; end if
          ) ; end begin
          (begin ; via
            (if (member layerNum viaLayerNumList)
              (begin
                (set! layerName (dbGetLayerNameFromNumber lib layerNum))
                (set! spacing (dbFetchLayerInfo lib layerNum "minSpacing"))

                (display "    contact_layer ( \"" file)
                (display layerName file)
                (display "\" );" file)
                (newline file)
              ) ;end begin
            )
          )
        ) ; end if
      )) ; end begin and lambda
      layerNumList
    ) ; end for-each
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpVia
;  Function: Dump Via section
;  Last Modifed: 9/7/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpVia (lambda (lib file)
  (let
    (
      (viaList (dbFetchContactNames))
      (contactCodeIdx  0)
      (lengthPrecision 0)
      (resPrecision    0)
      (ul              0)
      (ll              0)
      (cut             0)
      (cutName        "")
      (cutMaskName    "")
      (ulName         "")
      (llName         "")
      (width           0)
      (height          0)
      (ulEncWidth      0)
      (ulEncHeight     0)
      (llEncWidth      0)
      (llEncHeight     0)
      (res             0)
      (lx              0)
      (ly              0)
      (ux              0)
      (uy              0)
    )
    (set! lengthPrecision (dbFetchTechInfo lib "lengthPrecision"))
    (set! resPrecision (dbFetchTechInfo lib "resistancePrecision"))
    (for-each ( lambda (via)
      (begin
        (set! contactCodeIdx (dbFetchNonDefaultContactCodeInfo lib via "contactCodeNumber"))
        (set! cut (dbFetchNonDefaultContactCodeInfo lib via "cutLayerNumber"))
        (set! ul (dbFetchNonDefaultContactCodeInfo lib via "upperLayerNumber"))
        (set! ll (dbFetchNonDefaultContactCodeInfo lib via "lowerLayerNumber"))
        (set! cutName (dbGetLayerNameFromNumber lib cut))
        (set! cutMaskName (dbFetchLayerInfo lib cut "maskName"))
        (set! ulName (dbGetLayerNameFromNumber lib ul))
        (set! llName (dbGetLayerNameFromNumber lib ll))
        (set! width (dbFetchNonDefaultContactCodeInfo lib via "cutWidth"))
        (set! height (dbFetchNonDefaultContactCodeInfo lib via "cutHeight"))
        (set! ulEncWidth (dbFetchNonDefaultContactCodeInfo lib via "upperLayerEncWidth"))
        (set! ulEncHeight (dbFetchNonDefaultContactCodeInfo lib via "upperLayerEncHeight"))
        (set! llEncWidth (dbFetchNonDefaultContactCodeInfo lib via "lowerLayerEncWidth"))
        (set! llEncHeight (dbFetchNonDefaultContactCodeInfo lib via "lowerLayerEncHeight"))
        (set! res (dbFetchNonDefaultContactCodeInfo lib via "unitMaxResistance"))
        (set! res (/ res resPrecision))

        (if (not (equal? cutMaskName "polyCont"))
          (begin
            (display "    via ( \"" file)
            (display via file)
            (display "\" ) {" file)
            (newline file)
            (display "      is_default : TRUE;" file)
            (newline file)
            (display "      resistance : " file)
            (display res file)
            (display ";" file)
            (newline file)

            (display "      via_layer ( \"" file)
            (display ulName file)
            (display "\" ) {" file)
            (newline file)

            (set! lx (/ (- (- 0 (/ width 2)) ulEncWidth) lengthPrecision))
            (set! ly (/ (- (- 0 (/ height 2)) ulEncHeight) lengthPrecision))
            (set! ux (/ (+ (/ width 2) ulEncWidth) lengthPrecision))
            (set! uy (/ (+ (/ height 2) ulEncHeight) lengthPrecision))

            (display "        rectangle ( " file)
            (display lx file)(display ", " file)
            (display ly file)(display ", " file)
            (display ux file)(display ", " file)
            (display uy file)(display " );" file)
            (newline file)
            (display "      } /* end via_layer */" file)
            (newline file)

            (display "      via_layer ( \"" file)
            (display cutName file)
            (display "\" ) {" file)
            (newline file)

            (set! lx (/ (- 0 (/ width 2)) lengthPrecision))
            (set! ly (/ (- 0 (/ height 2)) lengthPrecision))
            (set! ux (/ (/ width 2) lengthPrecision))
            (set! uy (/ (/ width 2) lengthPrecision))

            (display "        rectangle ( " file)
            (display lx file)(display ", " file)
            (display ly file)(display ", " file)
            (display ux file)(display ", " file)
            (display uy file)(display " );" file)
            (newline file)
            (display "      } /* end via_layer */" file)
            (newline file)

            (display "      via_layer ( \"" file)
            (display llName file)
            (display "\" ) {" file)
            (newline file)

            (set! lx (/ (- (- 0 (/ width 2)) llEncWidth) lengthPrecision))
            (set! ly (/ (- (- 0 (/ height 2)) llEncHeight) lengthPrecision))
            (set! ux (/ (+ (/ width 2) llEncWidth) lengthPrecision))
            (set! uy (/ (+ (/ height 2) llEncHeight) lengthPrecision))

            (display "        rectangle ( " file)
            (display lx file)(display ", " file)
            (display ly file)(display ", " file)
            (display ux file)(display ", " file)
            (display uy file)(display " );" file)
            (newline file)

            (display "      } /* end via_layer */" file)
            (newline file)
            (display "    } /* end via */" file)
            (newline file)
          ) ; end begin
        ) ; end if
      )) ;end begin and lambda
      viaList
    ) ; end for-each
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpViaGen
;  Function: Dump via rule generate section
;  Last Modifed: 9/8/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpViaGen (lambda (lib file)
  (let
    (
      (hList (list "metal1" "metal3" "metal5" "metal7" "metal9"))
      (vList (list "metal2" "metal4" "metal6" "metal8" "metal10"))
      (viaMaskList (list "via1" "via2" "via3" "via4" "via5" "via6"
        "via7" "via8" "via9"))
      (viaList       (dbFetchContactNames))
      (contactCodeIdx  0)
      (lengthPrecision 0)
      (resPrecision    0)
      (ul              0)
      (ll              0)
      (cut             0)
      (cutName        "")
      (cutMaskName    "")
      (ulName         "")
      (llName         "")
      (width           0)
      (height          0)
      (ulEncWidth      0)
      (ulEncHeight     0)
      (llEncWidth      0)
      (llEncHeight     0)
      (overhang        0)
      (res             0)
      (lx              0)
      (ly              0)
      (ux              0)
      (uy              0)
      (maskName       "")
      (spacing         0)
      (dir            "")
    )
    (set! lengthPrecision (dbFetchTechInfo lib "lengthPrecision"))
    (set! resPrecision (dbFetchTechInfo lib "resistancePrecision"))
    (for-each ( lambda (via)
      (begin
        (set! contactCodeIdx (dbFetchNonDefaultContactCodeInfo lib via "contactCodeNumber"))
        (set! cut (dbFetchNonDefaultContactCodeInfo lib via "cutLayerNumber"))
        (set! ul (dbFetchNonDefaultContactCodeInfo lib via "upperLayerNumber"))
        (set! ll (dbFetchNonDefaultContactCodeInfo lib via "lowerLayerNumber"))
        (set! cutName (dbGetLayerNameFromNumber lib cut))
        (set! cutMaskName (dbFetchLayerInfo lib cut "maskName"))
        (set! ulName (dbGetLayerNameFromNumber lib ul))
        (set! llName (dbGetLayerNameFromNumber lib ll))
        (set! width (dbFetchNonDefaultContactCodeInfo lib via "cutWidth"))
        (set! height (dbFetchNonDefaultContactCodeInfo lib via "cutHeight"))
        (set! ulEncWidth (dbFetchNonDefaultContactCodeInfo lib via "upperLayerEncWidth"))
        (set! ulEncHeight (dbFetchNonDefaultContactCodeInfo lib via "upperLayerEncHeight"))
        (set! llEncWidth (dbFetchNonDefaultContactCodeInfo lib via "lowerLayerEncWidth"))
        (set! llEncHeight (dbFetchNonDefaultContactCodeInfo lib via "lowerLayerEncHeight"))
        (set! spacing (dbFetchNonDefaultContactCodeInfo lib via "minCutSpacing"))
        (set! spacing (/ spacing lengthPrecision))
        (set! res (dbFetchNonDefaultContactCodeInfo lib via "unitMaxResistance"))
        (set! res (/ res resPrecision))

        (if (not (equal? cutMaskName "polyCont"))
          (begin
            (display "    via_rule_generate ( \"" file)
            (display via file)
            (display "_array\" ) {" file)
            (newline file)

            (set! maskName (dbFetchLayerInfo lib ul "maskName"))
            (set! dir (dbFetchLayerInfo lib ul "preferredDirection"))
            (if (or (= dir 0) (null? dir))
              (cond
                ((member maskName hList)
                  (set! dir "horizontal")
                )
                ((member maskName vList)
                  (set! dir "vertical")
                )
              )
            )

            (display "      routing_formula ( \"" file)
            (display ulName file)
            (display "\" ) {" file)
            (newline file)
            (display "        routing_direction : " file)
            (display dir file)
            (display ";" file)
            (newline file)
            (display "        contact_overhang : " file)
            (display (/ ulEncWidth lengthPrecision) file)
            (display ";" file)
            (newline file)
            (display "        metal_overhang : 0.0 ;" file)
            (newline file)
            (display "      } /* end routing_formula */" file)
            (newline file)

            (set! maskName (dbFetchLayerInfo lib ll "maskName"))
            (set! dir (dbFetchLayerInfo lib ll "preferredDirection"))
            (if (or (= dir 0) (null? dir))
              (cond
                ((member maskName hList)
                  (set! dir "horizontal")
                )
                ((member maskName vList)
                  (set! dir "vertical")
                )
              )
            )

            (display "      routing_formula ( \"" file)
            (display llName file)
            (display "\" ) {" file)
            (newline file)
            (display "        routing_direction : " file)
            (display dir file)
            (display ";" file)
            (newline file)
            (display "        contact_overhang : " file)
            (display (/ llEncWidth lengthPrecision) file)
            (display ";" file)
            (newline file)
            (display "        metal_overhang : 0.0;" file)
            (newline file)
            (display "      } /* end routing_formula */" file)
            (newline file)


            (display "      contact_formula ( \"" file)
            (display cutName file)
            (display "\" ) {" file)
            (newline file)

            (set! lx (/ (- 0 (/ width 2)) lengthPrecision))
            (set! ly (/ (- 0 (/ height 2)) lengthPrecision))
            (set! ux (/ (/ width 2) lengthPrecision))
            (set! uy (/ (/ width 2) lengthPrecision))

            (display "        rectangle ( " file)
            (display lx file)(display ", " file)
            (display ly file)(display ", " file)
            (display ux file)(display ", " file)
            (display uy file)(display " );" file)
            (newline file)

            (display "        contact_spacing ( " file)
            (display (+ (/ width lengthPrecision) spacing) file)
            (display ", " file)
            (display (+ (/ height lengthPrecision) spacing) file)
            (display " );" file)
            (newline file)

            (display "        resistance : " file)
            (display res file)
            (display ";" file)
            (newline file)
            (display "      } /* end contact_formula */" file)
            (newline file)

            (display "    } /* end via_rule_generate */ " file)
            (newline file)
          ) ; end begin
        ) ; end if
      )) ;end begin and lambda
      viaList
    ) ; end for-each
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpSameNetSpacing
;  Function: Dump same net spacing section
;  Last Modifed: 10/26/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpSameNetSpacing (lambda (lib file)
  (let
    (
      (layerName  "")
      (maskName   "")
      (spacing     0)

    )
    (for-each (lambda (layer)
      (begin
        (set! layerName (dbGetLayerNameFromNumber lib layer))
        (set! maskName (dbFetchLayerInfo lib layer "maskName"))
        (if (not (equal? maskName "poly"))
          (begin
            (set! spacing (dbFetchLayerInfo lib layer "minSpacing"))
            (display "    same_net_min_spacing ( " file)
            (display layerName file)
            (display ", " file)
            (display layerName file)
            (display ", " file)
            (display spacing file)
            (display ", false );" file)
            (newline file)
          )
        )
      ))
      routeLayerNumList
    )
    (for-each (lambda (layer)
      (begin
        (set! layerName (dbGetLayerNameFromNumber lib layer))
        (set! spacing (dbFetchLayerInfo lib layer "minSpacing"))
        (display "    contact_min_spacing ( " file)
        (display layerName file)
        (display ", " file)
        (display layerName file)
        (display ", " file)
        (display spacing file)
        (display " );" file)
        (newline file)
      ))
      viaLayerNumList
    )
    (newline file)
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpSite
;  Function: Dump site definition section
;  Last Modifed: 9/8/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpSite (lambda (lib file)
(call-with-current-continuation (lambda (ret)
  (let
    (
      (cell   '())
      (boundary 0)
      (bbox   '())
      (width    0)
      (size     0)
      (height   0)
    )
    (set! cell (dbOpenCell "unitTile.CEL" "r"))
    (if (or (equal? cell #f) (or (= cell 0) (null? cell)))
      (begin
        (display "WARNING: No site definition is extracted.")
        (newline)
      )
      (begin
        (set! boundary (dbFetchObjectField cell '() "prBoundaryObjId"))
        (if (or (= boundary 0) (null? boundary))
          (set! boundary (dbFetchObjectField cell '() "cellBoundaryObjId"))
        )
        (set! bbox (dbFetchObjectField cell boundary "bbox"))
        (if (null? bbox)
          (begin
            (display "WARNING: unitTile cell has no PR boundary and cell boundary.")
            (newline)
            (display "WARNING: No site definition is extracted.")
            (newline)
            (ret #f)
          )
        )
        (set! width (abs (- (caadr bbox) (caar bbox))))
        (set! height (abs (- (cadadr bbox) (cadar bbox))))

        (display "    site ( \"" file)
        (display plibSiteName file)
        (display "\" ) {" file)
        (newline file)
        (display "      site_class : core;" file)
        (newline file)
        (display "      symmetry : y;" file)
        (newline file)
        (display "      size ( " file)
        (display width file)
        (display ", " file)
        (display height file)
        (display " );" file)
        (newline file)
        (display "    } /* end site */ " file)
        (newline file)
      )
    )
  )
))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpLib
;  Function: Dump the pin, blockage of the .FRAM view of all the cells
;  Last Modifed: 9/22/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpLib (lambda (lib file)
  (let*
    (
      (cellName "")
      (viewName "")
      (cvId      0)
      (cvName   "")
      (num       0)
    )

    (display "Dump Cell (Macro) abstract")(newline)
    (db-foreach lib '() "cellName" cell
      (begin
        (set! cellName (dbFetchObjectField lib cell "cellName"))
        (set! viewName (dbFetchObjectField lib cell "viewName"))
        (if (equal? viewName ".FRAM")
          (begin
            (set! cvName (string-append cellName viewName))
            (set! cvId (dbOpenCell cvName "r"))
            (display "    Translating ")
            (display cvName)
            (display "...")
            (newline)
            (dumpFRAMToPLIB cvId file plibSiteName)
            (newline file)
            (dbCloseCell cvId)
            (set! num (+ num 1))
          ) ;end begin
        ) ;end if
        #t
      ) ; end begin
    ) ;end db-foreach

    (newline)
    (display num)
    (display " cells are translated successfully.")
    (newline)
  ) ;end let*
)) ;end lambda and define

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: dumpFRAMToPLIB
;  Function: Dump the abstract record of the specified cell into
;            a PLIB file
;  Last Modifed: 10/17/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define dumpFRAMToPLIB (lambda (cvId file site)
  (call-with-current-continuation (lambda (ret)
    (let*
      (
        (pinMaskList (list "metal1" "via1" "metal2" "via2" "metal3" "via3"
            "metal4" "via4" "metal5" "via5" "metal6" "via6" "metal7" "via7"
            "metal8" "via8" "metal9" "via9" "metal10")
        )
        (lib (dbGetLibIdByCellId cvId))
        (cellName "")
        (prBoundary '())
        (cellBBox '())
        (origin '())
        (cellFlag 0)
        (bBox '())
        (xsize 0)
        (ysize 0)
        (pinName "")
        (pinList '())
        (pinInfo '())
        (layerInfo '())
        (portId 0)
        (portType "signal")
        (dir 0)
        (direction "")
        (layer 0)
        (layerId 0)
        (layerNum 0)
        (layerName "")
        (maskName "")
        (net  0)
        (index 0)
        (len 0)
        (point '())
        (pointList '())
        (index 0)
        (num 0)
        (i 0)
        (j 0)
        (obsList '())
        (obsInfo '())
        (class "core")
        (symmetry "rxy")
        (count 0)
      )

      (if (equal? #t debugFlag)
        (begin
          (display
            "********************************************************************************"
            dumpPLIBLogFile
          )
          (newline dumpPLIBLogFile)
          (display "Cell Record:" dumpPLIBLogFile)
          (newline dumpPLIBLogFile)
          (display (dbFetchObject cvId '()) dumpPLIBLogFile)
          (newline dumpPLIBLogFile)
          (newline dumpPLIBLogFile)
        )
      )
      (set! cellName (dbFetchObjectField cvId '() "cellName"))
      (set! prBoundary (dbFetchObjectField cvId '() "prBoundaryObjId"))
      (if (or (null? prBoundary) (= 0 prBoundary))
        (set! cellBBox (dbFetchObjectField cvId '() "cellBBox"))
        (set! cellBBox (dbFetchObjectField cvId prBoundary "bbox"))
      )
      (if (null? cellBBox)
        (begin
          (display "ERROR: Cell ")
          (display cellName)
          (display " has no PR boundary and cell boundary.")
          (display " Cell is ignored.")
          (newline)
          (ret #f)
        )
      )
      (set! origin (car cellBBox))
      (set! cellFlag (dbFetchObjectField cvId '() "cellFlag"))
      (cond
          ((> (bit-and cellFlag dbbStandardCell) 0)
            (begin (set! class "core")(set! symmetry "xy"))
          )
          ((> (bit-and cellFlag dbbModuleCell) 0)
            (begin (set! class "core")(set! symmetry "xy"))
          )
          ((> (bit-and cellFlag dbbDoubleHeightStdCell) 0)
            (begin (set! class "core")(set! symmetry "xy"))
          )
          ((> (bit-and cellFlag dbbCoverCell) 0)
            (begin (set! class "core")(set! symmetry "xy"))
          )
          ((> (bit-and cellFlag dbbIOPadCell) 0)
            (begin (set! class "pad")(set! symmetry "rxy"))
          )
          ((> (bit-and cellFlag dbbCornerPadCell) 0)
            (begin (set! class "pad")(set! symmetry "rxy"))
          )
          ((> (bit-and cellFlag dbbMacroCell) 0)
            (begin (set! class "ring")(set! symmetry "rxy"))
          )
          (else
            (begin (set! class "core")(set! symmetry "xy"))
          )
      )
      ; Pin record
      (if (equal? #t debugFlag)
        (begin
          (display "Pin Record:" dumpPLIBLogFile)
          (newline dumpPLIBLogFile)
          (newline dumpPLIBLogFile)
        )
      )
      (db-foreach cvId '() "Pin" pin
        (begin
          (if (equal? #t debugFlag)
            (begin
              (display (dbFetchObject cvId pin) dumpPLIBLogFile)
              (newline dumpPLIBLogFile)
              (newline dumpPLIBLogFile)
            )
          )
          (set! pinInfo '())
          (set! layerInfo '())
          (set! direction "")
          (set! portType "signal")

          (set! layer (dbFetchObjectField cvId pin "layer"))
          (set! layerId (dbFetchLayerIdByNumber lib layer))
          (set! maskName (dbFetchLayerInfo lib layerId "maskName"))
          (if (member maskName pinMaskList)
          (begin
              (set! pinName (dbFetchObjectField cvId pin "name"))
              (set! pinName (car (string-split-by-space pinName)))
              (set! pinName (busNameRule pinName))
              (set! index (list-find-assoc-index pinName pinList))
              (if (equal? #f index)
                (begin
                  (set! pinInfo (append pinInfo (list pinName)))
                  (set! portId (dbFetchObjectField cvId pin "portId"))
                  (set! dir (dbFetchObjectField cvId pin "accessDirection"))
                  (cond
                    ((> (bit-and dir dbbPinInputPort) 0)
                      (set! direction "input")
                    )
                    ((> (bit-and dir dbbPinOutputPort) 0)
                      (set! direction "output")
                    )
                    ((> (bit-and dir dbbPinInputOutputPort) 0)
                      (set! direction "inout")
                    )
                    ((> (bit-and dir dbbPinTriState) 0)
                      (set! direction "tristate")
                    )
                    (else
	              (set! direction "inout")
                    )
                  ) ; end cond
                  (set! pinInfo (append pinInfo (list direction)))

                  (set! dir (dbFetchObjectField cvId portId "direction"))
                  (cond
                    ((= (bit-and dir dbbPortStatusMask) dbcPortIsGND)
                      (set! portType "ground")
                    )
                    ((= (bit-and dir dbbPortStatusMask) dbcPortIsVDD)
                      (set! portType "power")
                    )
                    (else
                      (set! portType "signal")
                    )
                  )
                  (set! pinInfo (append pinInfo (list portType)))

                  (set! layerNum (dbGetLayerNameFromNumber lib layer))
                  (set! layerInfo (append layerInfo (list layerNum)))

                  (set! pointList (dbFetchObjectField cvId pin "points"))
                  (set! len (length pointList))
                  (if (zero? len)
                    (begin
                      (set! len 2)
                      (set! pointList (dbFetchObjectField cvId pin "bbox"))
                    )
                  )
                  (set! layerInfo (append layerInfo (list pointList)))
                  (set! pinInfo (append pinInfo (list layerInfo)))
                  (set! pinList (append pinList (list pinInfo)))
                )
                (begin
                  (set! pinInfo (list-ref pinList index))
                  (set! layerNum (dbGetLayerNameFromNumber lib layer))
                  (set! layerInfo (append layerInfo (list layerNum)))

                  (set! pointList (dbFetchObjectField cvId pin "points"))
                  (set! len (length pointList))
                  (if (zero? len)
                    (begin
                      (set! len 2)
                      (set! pointList (dbFetchObjectField cvId pin "bbox"))
                    )
                  )
                  (set! layerInfo (append layerInfo (list pointList)))
                  (set! pinInfo (append pinInfo (list layerInfo)))
                  (set! pinList (list-replace-nth-element pinInfo index pinList))
                )
              ) ; end if
            ) ;end begin
          ) ; end if
          #t
        ) ; end begin
      ) ;end db-foreach

      ; Blockage record
      (if (equal? #t debugFlag)
        (begin
          (display "Blockage Record:" dumpPLIBLogFile)
          (newline dumpPLIBLogFile)
          (newline dumpPLIBLogFile)
        )
      )
      (db-foreach cvId '() "Rectangle" rect
        (begin
          (if (equal? #t debugFlag)
            (begin
              (display (dbFetchObject cvId rect) dumpPLIBLogFile)
              (newline dumpPLIBLogFile)
              (newline dumpPLIBLogFile)
            )
          )
          (set! obsInfo '())
          (set! index 0)
          (set! layer (dbFetchObjectField cvId rect "layer"))
          (set! layerId (dbFetchLayerIdByNumber lib layer))
          (set! maskName (dbFetchLayerInfo lib layerId "maskName"))
          (set! layerName (dbFetchLayerInfo lib layerId "name"))
          (set! net (dbFetchObjectField cvId rect "netId"))
          (if (null? net)
            (set! net 0)
          )
          (set! bBox (dbFetchObjectField cvId rect "bbox"))
          (if (and (= net 0) (member maskName pinMaskList))
            (begin
              (set! index (list-find-assoc-index layerName obsList))
              (if (equal? #f index)
                (begin
                  (set! obsInfo (append obsInfo (list layerName)))
                  (set! obsInfo (append obsInfo (list bBox)))
                  (set! obsList (append obsList (list obsInfo)))
                )
                (begin
                  (set! obsInfo (list-ref obsList index))
                  (set! obsInfo (append obsInfo (list bBox)))
                  (set! obsList (list-replace-nth-element obsInfo index obsList))
                )
              ) ; end if
            ) ; end begin
          ) ; end if
          #t
        ) ; end begin
      ) ; end db-foreach

      (display "  macro ( \"" file)
      (display cellName file)
      (display "\" ) {" file)
      (newline file)
      (display "    cell_type : " file)
      (display class file)
      (display ";" file)
      (newline file)
      (display "    source : user;" file)
      (newline file)
      (display "    origin ( " file)
      (display (car origin) file)
      (display ", " file)
      (display (cadr origin) file)
      (display " );" file)
      (newline file)
      (set! xsize (abs (- (caadr cellBBox) (car origin))))
      (set! ysize (abs (- (cadadr cellBBox) (cadr origin))))
      (display "    size ( " file)
      (display xsize file)
      (display ", " file)
      (display ysize file)
      (display " );" file)
      (newline file)
      (display "    symmetry : " file)
      (display symmetry file)
      (display ";" file)
      (newline file)
      (display "    in_site : " file)
      (display site file)
      (display ";" file)
      (newline file)
      (display "    foreign ( \"" file)
      (display cellName file)
      (display "\" ) {" file)
      (newline file)
      (display "      origin ( " file)
      (display (car origin) file)
      (display ", " file)
      (display (cadr origin) file)
      (display " );" file)
      (newline file)
      (display "    } /* end foreign */" file)
      (newline file)
      (newline file)

      (for-each
        (lambda (pin)
          (set! pinName (car pin))
          (display "    pin ( \"" file)
          (display pinName file)
          (display "\" ) {" file)
          (newline file)

          (set! direction (cadr pin))
          (display "      pin_dir : " file)
          (display direction file)
          (display ";" file)
          (newline file)

          (display "      pin_type : " file)
          (set! portType (caddr pin))
          (cond
            ((equal? portType "power")
              (begin
                (display "power;" file)
                (newline file)
              )
            )
            ((equal? portType "ground")
              (begin
                (display "ground;" file)
                (newline file)
              )
            )
            (else
              (begin
                (display "signal;" file)
                (newline file)
              )
            )
          )

          (display "      foreign ( \"" file)
          (display pinName file)
          (display "\" ) {" file)
          (newline file)
          (display "      } /* end foreign */" file)
          (newline file)
          (newline file)

          (set! len (length pin))
          (do
            ((i 3 (+ i 1)))
            ((equal? i len))
            (display "      port () {" file)
            (newline file)
            (display "        geometry ( \"" file)
            (display (car (list-ref pin i)) file)
            (display "\" ) {" file)
            (newline file)
            (display "          rectangle ( " file)
            (set! pointList (cadr (list-ref pin i)))
            (set! num (length pointList))
            (set! count 0)
            (do
              ((j num (- j 1)))
              ((zero? j))
              (set! point (car pointList))
              (set! pointList (cdr pointList))
              (set! count (+ count 1))
              (display (car point) file)
              (display ", " file)
              (set! count (+ count 1))
              (display (cadr point) file)
              (if (= count 4)
                (set! count 0)
                (display ", " file)
              )
              (if (and (odd? j) (> j 1))
                (begin
                  (display " );" file)
                  (newline file)
                  (display "          rectangle ( " file)
                )
              )
              #t
            )
            (display " );" file)
            (newline file)
            (display "        } /* end geometry */" file)
            (newline file)
            (display "      } /* end port */" file)
            (newline file)
            #t
          )
          ;(display "      } /* end port */" file)
          ;(newline file)
          (display "    } /* end pin */" file)
          (newline file)
          (newline file)
          #t
        )
        pinList
      ) ; end for-each

      (display "    obs () {" file)
      (newline file)
      (for-each
        (lambda (obs)
          (set! len (length obs))
          (set! layerName (car obs))

          (display "      geometry ( \"" file)
          (display layerName file)
          (display "\" ) {" file)
          (newline file)

          (do
            ((i 1 (+ i 1)))
            ((equal? i len))
            (begin
              (set! bBox (list-ref obs i))
              (display "        rectangle ( " file)
              (display (caar bBox) file)
              (display ", " file)
              (display (cadar bBox) file)
              (display ", " file)
              (display (caadr bBox) file)
              (display ", " file)
              (display (cadadr bBox) file)
              (display " );" file)
              (newline file)
            ) ; end begin
          ) ; end do
          (display "      } /* end geometry */" file)
          (newline file)
          (newline file)
          #t
        ) ; end lambda
        obsList
      ) ; end for-each
      (display "    } /* end obs */" file)
      (newline file)
      (display "  } /* end macro */" file)
      (newline file)
      #t
    ) ;end let*
  )) ; end call and lambda
)) ;end lambda and define


(define string-split-by-space (lambda (s)
(let
  ((result '()) (sl (string-length s)))
  (if (zero? sl)
    result
    (let loop ((s s) (sp sl) (r result))
      (let
        ((nwp (string-skip-white s sp)))
        (if (zero? nwp)
          r
          (let*
            ((nsp (string-skip-non-white s nwp))
            (nr (cons (substring s nsp nwp) r)))
            (if (zero? nsp) nr (loop s (- nsp 1) nr))
          )
        )
      )
    )
  )
)
))

(define string-skip-white (lambda (s i)
(cond
  ((zero? i) i)
  ((char-whitespace? (string-ref s (- i 1))) (string-skip-white s (- i 1)))
  (else i)
)
))

(define string-skip-non-white (lambda (s i)
(cond
  ((zero? i) i)
  ((char-whitespace? (string-ref s (- i 1))) i)
  (else (string-skip-non-white s (- i 1)))
)
))


(define list-find-assoc-index (lambda (name infoList)
(let
  (
    (index #f)
    (len (length infoList))
    (i 0)
  )
  (do
    ((i 0 (+ i 1)))
    ((equal? len i))
    (if (equal? name (car (list-ref infoList i)))
      (set! index i)
    )
  )
  index
)
))


(define list-replace-nth-element (lambda (a n aList)
(let
  (
    (len (length aList))
    (bList '())
    (i 0)
  )
  (do
    ((i 0 (+ i 1)))
    ((equal? n i))
    (set! bList (append bList (list (list-ref aList i))))
  )
  (set! bList (append bList (list a)))
  (set! bList (append bList (cdr (list-tail aList n))))
  bList
)
))

(define replace-string-char (lambda (str char newChar)
  (let
    (
      (newStr "")
      (len 0)
    )
    (set! newStr (string-copy str))
    (set! len (string-length str))
    (do
      ((i 0 (+ i 1)))
      ((equal? i len))
      (if (equal? char (string-ref str i))
        (string-set! newStr i newChar)
      )
    )
    newStr
  )
))


(define busNameRule (lambda (name)
  (let
    ((newName ""))
    (set! newName (string-copy name))
    (cond
      ((equal? plibBusNameRule "[%d]")
        (set! newName (replace-string-char newName #\( #\[))
        (set! newName (replace-string-char newName #\) #\]))
        (set! newName (replace-string-char newName #\< #\[))
        (set! newName (replace-string-char newName #\> #\]))
        (set! newName (replace-string-char newName #\{ #\[))
        (set! newName (replace-string-char newName #\} #\]))
      )
      ((equal? plibBusNameRule "<%d>")
        (set! newName (replace-string-char newName #\( #\<))
        (set! newName (replace-string-char newName #\) #\>))
        (set! newName (replace-string-char newName #\[ #\<))
        (set! newName (replace-string-char newName #\] #\>))
        (set! newName (replace-string-char newName #\{ #\<))
        (set! newName (replace-string-char newName #\} #\>))
      )
      ((equal? plibBusNameRule "(%d)")
        (set! newName (replace-string-char newName #\[ #\())
        (set! newName (replace-string-char newName #\] #\)))
        (set! newName (replace-string-char newName #\< #\())
        (set! newName (replace-string-char newName #\> #\)))
        (set! newName (replace-string-char newName #\{ #\())
        (set! newName (replace-string-char newName #\} #\)))
      )
      ((equal? plibBusNameRule "{%d}")
        (set! newName (replace-string-char newName #\( #\{))
        (set! newName (replace-string-char newName #\) #\}))
        (set! newName (replace-string-char newName #\< #\{))
        (set! newName (replace-string-char newName #\> #\}))
        (set! newName (replace-string-char newName #\[ #\{))
        (set! newName (replace-string-char newName #\] #\}))
      )
    )
    newName
  )
))

