
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: readPDEF3
;  Function: Read the design from the Physical Compiler's
;            output PDEF (IEEE PDEF) into Apollo
;  Last Modifed: 8/17/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define readPDEF3 (lambda (filename)
  (let
    (
      (cv (geGetEditCell))
    )
    (define  pdefLengthResolution 0)
    (define  pdefList     '())
    (define  layerList    '())
    (define  viaList      '())
    (define  trackList    '())
    (define  cellList     '())
    (define  pinList      '())
    (define  rowList      '())
    (define  pnetList     '())
    (define  netList      '())
    (define  blockageList '())
    (define  placeObsList '())
    (define  routeObsList '())
    (define  boundList    '())
    (define  cellBoundary '())
    (define  coreBoundary '())

    (display "Parsing PDEF file...")(newline)
    (set! pdefList (cdr (with-input-from-file filename read)))

    (for-each (lambda (item)
      (begin
        (cond
          ((equal? 'CLUSTER (car item))
            (for-each (lambda (obj)
              (begin
                (case (car obj)
                  ('PIN
                    (set! pinList (cons obj pinList))
                  )
                  ('CELL
                    (set! cellList (cons obj cellList))
                  )
                  ('PNET
                    (set! pnetList (cons obj pnetList))
                  )
                  ('NET
                    (set! netList (cons obj netList))
                  )
                  ('CLUSTER
                    (if (equal? 'OBSTRUCTION (caaddr obj))
                      (set! placeObsList (cons obj placeObsList))
                      (set! boundList (cons obj boundList))
                    )
                  )
                  ('OBSTRUCTION
                    (set! routeObsList (cons obj routeObsList))
                  )
                )
              ))
              (cddr item)
            )
          )
          ((equal? 'SITE (car item))
            (set! rowList (cons item rowList))
          )
          ((equal? 'LAYER_DEF (car item))
            (set! layerList (cdr item))
          )
          ((equal? 'VIA_DEF (car item))
            (set! viaList (cdr item))
          )
          ((equal? 'DISTANCE_UNIT (car item))
            (set! pdefLengthResolution (cadr item))
          )
          ((equal? 'DEF_DIEAREA (car item))
            (set! cellBoundary
              (list
                (* pdefLengthResolution (string->number (cadr item)))
                (* pdefLengthResolution (string->number (caddr item)))
                (* pdefLengthResolution (string->number (cadddr item)))
                (* pdefLengthResolution (string->number (list-ref item 4)))
              )
            )
          )
          ((equal? 'CORE_AREA (car item))
            (set! coreBoundary
              (list
                (* pdefLengthResolution (cadr item))
                (* pdefLengthResolution (caddr item))
                (* pdefLengthResolution (cadddr item))
                (* pdefLengthResolution (list-ref item 4))
              )
            )
          )
          ((if
            (and
              (> (string-length (symbol->string (car item))) 9)
              (equal? "DEF_TRACK" (substring (symbol->string (car item)) 0 9))
            )
            (set! trackList (cons item trackList))
          ))
        ) ; end case
      )) ; end begin and lambda
      pdefList
    ) ; end for-each
    (if (null? cellBoundary)
      (begin
        (display "ERROR: Missing cell boundary.")(newline)
      )
      (begin
        (dbCreateCellBoundary cv
          (list
            (list (car cellBoundary) (cadr cellBoundary))
            (list (caddr cellBoundary) (cadr cellBoundary))
            (list (caddr cellBoundary) (cadddr cellBoundary))
            (list (car cellBoundary) (cadddr cellBoundary))
            (list (car cellBoundary) (cadr cellBoundary))
          )
        )
      )
    )
    (if (null? coreBoundary)
      (begin
        (display "ERROR: Missing core boundary.")(newline)
      )
    )
    (display "Read IO pin information...")(newline)
    (readPin cv cellBoundary pinList)
    (display "Read cell row information...")(newline)
    (readCellRow cv coreBoundary rowList)
    (display "Read wire track information...")(newline)
    (readTrack cv cellBoundary layerList trackList)
    (display "Read cell placement information...")(newline)
    (readCellPlacement cv cellList)
    (display "Read group and region information...")(newline)
    (readGroupRegion cv boundList)
    (display "Read blockage information...")(newline)
    (readPlacementBlockage cv placeObsList)
    (axLoadDesignSetupDone)
    (display "Read special net routing information...")(newline)
    (readPNET cv layerList viaList pnetList)
    (display "Read regular net routing information...")(newline)
    (readPNET cv layerList viaList netList)
    #t
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: readPin
;  Function: Read the pin information
;  Last Modifed: 8/17/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define readPin (lambda (cv boundary plist)
  (let
    (
      (lib (dbGetLibIdByCellId cv))
      (pinName  "")
      (pinNumber '())
      (attrList '())
      (x0 0)
      (y0 0)
      (x1 0)
      (y1 0)
      (pin_type "")
      (pin_dir "")
      (pin_info '())
      (pin_layer "")
      (bbox '())
      (net_name "")
      (layerNum 0)
      (accessDirection   0)
      (statusFlag 1)
    )
    (for-each (lambda (pin)
      (begin
        (set! pinName (symbol->string (cadr pin)))
        (set! attrList (cddr pin))
        (for-each (lambda (attr)
        (begin
          (cond
            ((equal? 'PIN_NUMBER (car attr))
              (begin
                (set! pinNumber (number->string (cadr attr)))
                (set! pinName (string-append pinName "\ "))
                (set! pinName (string-append pinName pinNumber))
              )
            )
            ((equal? 'LOC (car attr))
              (begin
                (set! x0 (* pdefLengthResolution (cadr attr)))
                (set! y0 (* pdefLengthResolution (caddr attr)))
              )
            )
            ((equal? 'PLACED (car attr))
              (set! statusFlag dbbPRPinPlaced)
            )
            ((equal? 'FIXED (car attr))
              (set! statusFlag dbbPRPinFixed)
            )
            ((equal? 'PIN_TYPE (car attr))
              (set! pin_type (symbol->string (cadr attr)))
            )
            ((equal? 'PIN_DIR (car attr))
              (cond
                ((equal? 'INPUT (cadr attr))
                  (set! pin_dir dbbPinInputPort)
                )
                ((equal? 'OUTPUT (cadr attr))
                  (set! pin_dir dbbPinOutputPort)
                )
                ((equal? 'INOUT (cadr attr))
                  (set! pin_dir dbbPinInputOutputPort)
                )
              ) ; end case
            )
            ((equal? 'PIN_LAYER (car attr))
              (begin
                (set! pin_info (string-parse (cadr attr)))
                (set! pin_layer (car pin_info))
                (set! bbox
                  (list
                    (* pdefLengthResolution (list-ref pin_info 2))
                    (* pdefLengthResolution (list-ref pin_info 3))
                    (* pdefLengthResolution (list-ref pin_info 6))
                    (* pdefLengthResolution (list-ref pin_info 7))
                  )
                ) ; end set!
              )
            )
            ((equal? 'DEF_NET_NAME (car attr))
              (set! net_name (cadr attr))
            )
          ))) ; case and lambda
          attrList
        ) ; end for-each
        (set! layerNum (dbGetLayerNumberFromName lib pin_layer))
        (set! x0 (+ x0 (car bbox)))
        (set! y0 (+ y0 (cadr bbox)))
        (set! x1 (+ x0 (caddr bbox)))
        (set! y1 (+ y0 (cadddr bbox)))
        (cond
          ((= x0 0.0)
            (set! accessDirection dbbLeftAccessDir)
          )
          ((= y0 0.0)
            (set! accessDirection dbbDownAccessDir)
          )
          ((= x1 (caddr boundary))
            (set! accessDirection dbbRightAccessDir)
          )
          ((= y1 (cadddr boundary))
            (set! accessDirection dbbUpAccessDir)
          )
        )
        (set! accessDirection (bit-or accessDirection pin_dir))
        (dbRestorePin cv net_name pinName statusFlag accessDirection
          0 0 layerNum (list (list x0 y0) (list x1 y1))
        )
      )) ; begin and lambda
      plist
    ) ; end for-each
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: readCellRow
;  Function: Read the cell row section
;  Last Modifed: 8/14/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define readCellRow (lambda (cv bbox rlist)
  (let
    (
      (unitTile   0)
      (boundary   0)
      (tileBBox '())
      (tileHeight 0)
      (tileWidth  0)
      (num        0)
      (ll       '())
      (x0         0)
      (y0         0)
      (x1         0)
      (y1         0)
    )

    (set! unitTile (dbOpenCell "unitTile.CEL" "r"))
    (if (or (null? unitTile) (= 0 unitTile))
      (begin
        (display "ERROR Cannot find unitTile cell in the reference library.")
        (newline)
      )
      (begin
        (set! boundary (dbFetchObjectField unitTile '() "prBoundaryObjId"))
        (if (or (null? boundary) (= 0 boundary))
          (set! tileBBox (dbFetchObjectField unitTile '() "cellBBox"))
          (set! tileBBox (dbFetchObjectField unitTile boundary "bbox"))
        )
        (set! tileWidth (abs (- (caadr tileBBox) (caar tileBBox))))
        (set! tileHeight (abs (- (cadadr tileBBox) (cadar tileBBox))))
      )
    )
    (dbCloseCell unitTile)

    (axPurgeSingleRecordType cv "tile pattern")
    (axPurgeSingleRecordType cv "row")
    (axPurgeSingleRecordType cv "base array")
    (axCreateBaseArrayRecord cv "unit" #t
      (list (car bbox) (cadr bbox))
      (list (caddr bbox) (cadddr bbox))
    )
    (axCreateTilePatternRecord cv "unit" "unit" #t #f)
    (axCreateTilePatternRecord cv "unit" "unit-flip" #t #t)
    (axCreateTilePatternRecord cv "unit" "unit-v" #f #f)
    (axCreateTilePatternRecord cv "unit" "unit-flip-v" #f #t)

    (for-each (lambda (row)
      (begin
        (set! ll (string-parse (list-ref row 3)))
        (set! x0 (* pdefLengthResolution (car ll)))
        (set! y0 (* pdefLengthResolution (cadr ll)))
        ;(set! tileWidth (* pdefLengthResolution (string->number (list-ref row 6))))
        (set! num (string->number (list-ref row 7)))
        (set! x1 (+ x0 (* num tileWidth)))
        (set! y1 (+ y0 tileHeight))

        (if (equal? "0" (list-ref row 5))
          (axCreateRowRecord cv #t #f "unit" (list x0 y0) (list x1 y1))
          (axCreateRowRecord cv #t #t "unit-flip" (list x0 y0) (list x1 y1))
        )
      ))
      rlist
    )
  )
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: readTrack
;  Function: Read the cell row section
;  Last Modifed: 8/18/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define readTrack (lambda (cv bbox llist tlist)
  (let*
    (
      (lib (dbGetLibIdByCellId cv))
      (layerNum    0)
      (layerName  "")
      (layerList '())
      (dir        "")
      (start       0)
      (step        0)
      (count       0)
      (maskList  '())
      (maskName   "")
      (layerDir   #t)
      (x0          0)
      (y0          0)
      (x1          0)
      (y1          0)
    )

    (axPurgeSingleRecordType cv "track")
    (axPurgeSingleRecordType cv "wire dir")

    (for-each (lambda (layer)
      (begin
        (set! layerNum (caddr layer))
        (set! maskName (dbFetchLayerInfo lib layerNum "maskName"))
        (cond
          ((equal? maskName "metal1")
            (axCreateWireDirRecord cv layerNum #t)
          )
          ((equal? maskName "metal2")
            (axCreateWireDirRecord cv layerNum #f)
          )
          ((equal? maskName "metal3")
            (axCreateWireDirRecord cv layerNum #t)
          )
          ((equal? maskName "metal4")
            (axCreateWireDirRecord cv layerNum #f)
          )
          ((equal? maskName "metal5")
            (axCreateWireDirRecord cv layerNum #t)
          )
          ((equal? maskName "metal6")
            (axCreateWireDirRecord cv layerNum #f)
          )
          ((equal? maskName "metal7")
            (axCreateWireDirRecord cv layerNum #t)
          )
        ) ; end cond
      )) ; end begin and lambda
      llist
    ) ; end for-each

    (for-each (lambda (track)
      (begin
        (set! dir (cadr track))
        (set! start (* pdefLengthResolution (string->number (caddr track))))
        (set! count (string->number (list-ref track 3)))
        (set! step (* pdefLengthResolution (string->number (list-ref track 4))))
        (set! layerList (string-parse (list-ref track 5)))

        (for-each (lambda (layer)
          (begin
            (set! layerNum (dbGetLayerNumberFromName lib layer))
            (cond
              ((equal? dir "Y") ; horizontal track
                (set! x0 0.0)
                (set! y0 start)
                (set! x1 (caddr bbox))
                (set! y1 (+ start (* step (- count 1))))
                (axCreateTrackRecord cv layerNum #t start step count
                  (list x0 y0) (list x1 y1)
                )
              )
              ((equal? dir "X") ; vertical track
                (set! x0 start)
                (set! y0 0.0)
                (set! x1 (+ start (* step (- count 1))))
                (set! y1 (cadddr bbox))
                (axCreateTrackRecord cv layerNum #f start step count
                  (list x0 y0) (list x1 y1)
                )
              )
              (else
                (display "ERROR: Unknown track direction ")
                (display dir)
                (newline)
              )
            ) ; end cond
          ))
          layerList
        )
      ))
      tlist
    )
    #t
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: readCellPlacement
;  Function: Read the placement section
;  Last Modifed: 9/17/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define readCellPlacement (lambda (cv clist)
  (let
    (
      (cellName  "")
      (attrList '())
      (llx      0.0)
      (lly      0.0)
      (orient    "")
      (inst       0)
      (master     0)
      (bbox     '())
      (width      0)
      (height     0)
      (dx         0)
      (dy         0)
      (fix       #f)
    )

    (dbSetCurrentRegionNumber cv 0)
    (for-each (lambda (cell)
      (begin
        (set! cellName (symbol->string (cadr cell)))
        (set! attrList (cddr cell))
        (set! fix #f)
        (for-each (lambda (attr)
          (cond
            ((equal? 'LOC (car attr))
              (begin
                (set! llx (* pdefLengthResolution (cadr attr)))
                (set! lly (* pdefLengthResolution (caddr attr)))
              )
            )
            ((equal? 'ORIENT (car attr))
              (set! orient (cadr attr))
            )
            ((equal? 'FIXED (car attr))
              (set! fix #t)
            )
          ))
          attrList
        )
        (set! inst (dbGetCellInstByName cv cellName))
        (if inst
          (begin
            (set! master (dbFetchObjectField cv inst "cellInstMasterId"))
            (if (= 0 (dbFetchObjectField cv master "prBoundaryId"))
              (set! bbox (dbFetchObjectField cv master "cellBoundaryBBox"))
              (set! bbox (dbFetchObjectField cv master "prBoundaryBBox"))
            )
            (set! width (- (caadr bbox) (caar bbox)))
            (set! height (- (cadadr bbox) (cadar bbox)))
            (set! dx (- 0 (caar bbox)))
            (set! dy (- 0 (cadar bbox)))
            (cond
              ((equal? orient "0")
                (dbSetCellInstPlacement cv cellName "0" "no" "origin"
                  (list (+ llx dx) (+ lly dy))
                )
              )
              ((equal? orient "0-mirror")
                (dbSetCellInstPlacement cv cellName "0" "X" "origin"
                  (list (+ (- llx dx) width) (+ lly dy))
                )
              )
              ((equal? orient "90")
                (dbSetCellInstPlacement cv cellName "90" "no" "origin"
                  (list (+ (- llx dy) height) (+ lly dx))
                )
              )
              ((equal? orient "90-mirror")
                (dbSetCellInstPlacement cv cellName "90" "X" "origin"
                  (list (+ llx dy) (+ lly dx))
                )
              )
              ((equal? orient "180")
                (dbSetCellInstPlacement cv cellName "180" "no" "origin"
                  (list (+ (- llx dx) width) (+ (- lly dy) height))
                )
              )
              ((equal? orient "180-mirror")
                (dbSetCellInstPlacement cv cellName "180" "X" "origin"
                  (list (+ llx dx) (+ (- lly dy) height))
                )
              )
              ((equal? orient "270")
                (dbSetCellInstPlacement cv cellName "270" "no" "origin"
                  (list (+ llx dy) (+ (- lly dx) width))
                )
              )
              ((equal? orient "270-mirror")
                (dbSetCellInstPlacement cv cellName "270" "X" "origin"
                  (list (+ (- llx dy) height) (+ (- lly dx) width))
                )
              )
            ) ; end case
            (if (equal? #t fix)
              (dbSetCellInstStatus cv inst "Fixed")
              (dbSetCellInstStatus cv inst "Placed")
            )
          ) ; end begin
          (begin
            (display "ERROR: Cell ")
            (display cellName)
            (display " is not found.")
            (newline)
          ) ; end begin
        ) ; end if
      )) ; end begin and lambda
      clist
    ) ; end for-each
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: readGroupRegion
;  Function: Read the move bound (i.e. group or region) section
;  Last Modifed: 9/17/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define readGroupRegion (lambda (cv blist)
  (let
    (
      (name       "")
      (isRegion   #f)
      (objList   '())
      (llx         0)
      (lly         0)
      (urx         0)
      (ury         0)
    )

    (for-each (lambda (cluster)
      (begin
        (set! name (cadr cluster))
        (axCreateGroup cv name)
        (set! objList (cddr cluster))
        (set! isRegion #f)
        (for-each (lambda (obj)
          (begin
            (cond
              ((equal? 'RECT (car obj))
                (begin
                  (set! isRegion #t)
                  (set! llx (cadr obj))
                  (set! lly (caddr obj))
                  (set! urx (cadddr obj))
                  (set! ury (car (cddddr obj)))
                )
              )
              ((equal? 'CELL (car obj))
                (axAddCellToGroup cv name (symbol->string (cadr obj)))
              )
            ) ; end cond
          ))
          objList
        ) ; end for-each
        (if (equal? #t isRegion)
          (axCreateRegion cv "H" name
            (list (* pdefLengthResolution llx) (* pdefLengthResolution lly))
            (list (* pdefLengthResolution urx) (* pdefLengthResolution ury))
          ) ; "H" means cell row direction is horizontal
        )
      ))
      blist
    )
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: readBlockage
;  Function: Read the physical net section
;  Last Modifed: 8/18/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define readPlacementBlockage (lambda (cv blist)
  (let
    (
      (obsLayer  221) ; default blockage layer number in Avanti
      (attrList  '())
      (llx         0)
      (lly         0)
      (urx         0)
      (ury         0)
    )
    (axPurgePlaceBlockage cv)

    (for-each (lambda (obs)
      (begin
        (set! attrList (cddr obs))
        (for-each (lambda (attr)
          (case (car attr)
            ('RECT
              (set! llx (* pdefLengthResolution (cadr attr)))
              (set! lly (* pdefLengthResolution(caddr attr)))
              (set! urx (* pdefLengthResolution(cadddr attr)))
              (set! ury (* pdefLengthResolution(car (cddddr attr))))
            )
          ))
          attrList
        )
        (dbCreateRectangle cv obsLayer (list (list llx lly) (list urx ury)))
      ))
      blist
    )
  )
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: readPNET
;  Function: Read the physical net section
;  Last Modifed: 9/17/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define readPNET (lambda (cv llist vlist plist)
  (let
    (
      (layerName  "")
      (netName    "")
      (netType    "")
      (routeList '())
      (attrList  '())
      (routeType   0)
      (layerNum    0)
      (viaNum      0)
      (viaName    "")
      (nameList  '())
      (xDuplicate  0)
      (yDuplicate  0)
      (xPitch      0)
      (yPitch      0)
      (isVia      #f)
      (isVertical  0)
      (wireLength  0)
      (width       0)
      (x0          0)
      (y0          0)
      (x1          0)
      (y1          0)
      (pre_x       0)
      (pre_y       0)
      (viaMaster '())
      (index       0)
      (lib (dbGetLibIdByCellId cv))
    )

    (for-each (lambda (pnet)
      (begin
        (set! netName (symbol->string (cadr pnet)))
        (set! netType (cadr (caddr pnet)))
        (case netType
          ('pwr
            (set! routeType dbcPGRouteType)
          )
          ('gnd
            (set! routeType dbcPGRouteType)
          )
          ('data
            (set! routeType dbcSignalUserRouteType)
          )
          ('data
            (set! routeType dbcClockRouteType)
          )
        )
        (set! routeList (cdddr pnet))
        (for-each (lambda (route)
          (begin
            (set! attrList (cdr route))
            (for-each (lambda (attr)
              (cond
                ((equal? (car attr) 'LAYER_WIDTH)
                  (begin
                    (set! layerNum (cadr attr))
                    (set! width (* pdefLengthResolution (caddr attr)))
                  )
                )
                ((equal? (car attr) 'SHAPE)
                  (if
                    (and (= routeType dbcPGRouteType)
                      (equal? 'STRIPE (cadr attr))
                    )
                    (set! routeType dbcPGStrapRouteType)
                  )
                )
                ((equal? (car attr) 'VIA)
                  (begin
                    (set! index (cadr attr))
                    (set! x0 (caaddr attr))
                    (set! y0 (cadr (caddr attr)))
                    (set! isVia #t)
                    (if (equal? x0 '*)
                      (set! x0 pre_x)
                    )
                    (if (equal? y0 '*)
                      (set! y0 pre_y)
                    )
                  )
                )
                (else
                  (begin
                    (set! isVia #f)
                    (set! x0 (caadr attr))
                    (set! y0 (cadadr attr))
                    (if (equal? x0 '*)
                      (set! x0 pre_x)
                      (set! pre_x x0)
                    )
                    (if (equal? y0 '*)
                      (set! y0 pre_y)
                      (set! pre_y y0)
                    )
                    (if (equal? '* (caaddr attr))
                      (set! x1 x0)
                      (begin
                        (set! x1 (caaddr attr))
                        (set! pre_x x1)
                      )
                    )
                    (if (equal? '* (cadr (caddr attr)))
                      (set! y1 y0)
                      (begin
                        (set! y1 (cadr (caddr attr)))
                        (set! pre_y y1)
                      )
                    )
                  ) ; end begin
                ) ; end else
              )) ; end cond and lambda
              attrList
            ) ; end for-each

            (if (equal? #f isVia)
              (begin ; wire
                (if (= x1 x0)
                  (set! isVertical 1)
                )
                (if (= y1 y0)
                  (set! isVertical 0)
                )
                (if (= 0 isVertical)
                  (set! wireLength (abs (- x1 x0)))
                  (set! wireLength (abs (- y1 y0)))
                )
                (axCreateWire cv netName layerNum 0 0 width routeType
                  isVertical (* pdefLengthResolution wireLength)
                  (list
                    (* pdefLengthResolution x0)
                    (* pdefLengthResolution y0)
                  )
                )
              ) ; end begin (wire)
              (begin ; contact
                (set! viaMaster (findViaMaster vlist index))
                (if (null? viaMaster)
                  (begin
                    (display "ERROR: Cannot find via master ")
                    (display index)
                    (display " in the PDEF via table.")
                    (newline)
                  )
                  (begin
                    (set! viaName (cadr viaMaster))
                    (set! nameList (string-split-by-char viaName #\_))
                    (set! viaName (car nameList))
                    (set! viaNum
                      (dbFetchNonDefaultContactCodeInfo lib viaName "contactCodeNumber")
                    )
                    (if (= 1 (length nameList))
                      (axCreateContact cv netName viaNum 0 routeType
                        (list
                          (* pdefLengthResolution x0)
                          (* pdefLengthResolution y0)
                        )
                      )
                      (begin ; via array
                        (set! xDuplicate (string->number (cadr nameList)))
                        (set! yDuplicate (string->number (caddr nameList)))
                        (set! xPitch (string->number (cadddr nameList)))
                        (set! yPitch (string->number (car (cddddr nameList))))
                        (axCreateContactArray cv netName viaNum 0 routeType
                          (list
                            (* pdefLengthResolution x0)
                            (* pdefLengthResolution y0)
                          )
                          xPitch yPitch xDuplicate yDuplicate '()
                        )
                      )
                    ) ; end if
                  ) ; end begin
                ) ; end if
              ) ; end begin
            ) ; end if
          )) ; end begin and lambda
          routeList
        ) ; end for-each
      )) ; end begin and lambda
      plist
    ) ; end for-each
    #t
  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  Procedure: readNET
;  Function: Read the net section
;  Last Modifed: 9/18/00
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define readNET (lambda (cv llist vlist nlist)
  (let
    (
      (layerName  "")
      (netName    "")
      (netType    "")
      (routeList '())
      (attrList  '())
      (routeType   0)
      (layerNum    0)
      (viaNum      0)
      (viaName    "")
      (nameList  '())
      (xDuplicate  0)
      (yDuplicate  0)
      (xPitch      0)
      (yPitch      0)
      (isVertical  0)
      (wireLength  0)
      (width       0)
      (x0          0)
      (y0          0)
      (x1          0)
      (y1          0)
      (pre_x       0)
      (pre_y       0)
      (viaMaster '())
      (index       0)
      (first      #t)
      (lib (dbGetLibIdByCellId cv))
    )

    (for-each (lambda (net)
      (begin
        (set! netName (symbol->string (cadr net)))
        (set! netType dbcSignalUserRouteType)
        (set! routeList (cddr net))
        (if (equal? '() routeList)
          (begin
            (display "WARNING: Net ")
            (display netName)
            (display " has no physical route.")
            (newline)
          )
        )
        (for-each (lambda (route)
          (begin
            (set! attrList (cdr route))
            (set! first #t)
            (for-each (lambda (attr)
              (cond
                ((equal? (car attr) 'VIA)
                  (begin
                    (set! index (cadr attr))
                    (set! x0 (caaddr attr))
                    (set! y0 (cadr (caddr attr)))
                    (if (equal? first #t)
                      (begin
                        (set! pre_x x0)
                        (set! pre_y y0)
                        (set! first #f)
                      )
                      (begin
                        (if (equal? x0 '*)
                          (set! x0 pre_x)
                          (set! pre_x x0)
                        )
                        (if (equal? y0 '*)
                          (set! y0 pre_y)
                          (set! pre_y y0)
                        )
                      )
                    )
                    (set! viaMaster (findViaMaster vlist index))
                    (if (null? viaMaster)
                      (begin
                        (display "ERROR: Cannot find via master ")
                        (display index)
                        (display " in the PDEF via table.")
                        (newline)
                      )
                      (begin
                        (set! viaName (cadr viaMaster))
                        (set! nameList (string-split-by-char viaName #\_))
                        (set! viaName (car nameList))
                        (set! viaNum
                          (dbFetchNonDefaultContactCodeInfo lib viaName "contactCodeNumber")
                        )
                        (if (= 1 (length nameList))
                          (axCreateContact cv netName viaNum 0 routeType
                            (list
                              (* pdefLengthResolution x0)
                              (* pdefLengthResolution y0)
                            )
                          )
                          (begin ; via array
                            (set! xDuplicate (string->number (cadr nameList)))
                            (set! yDuplicate (string->number (caddr nameList)))
                            (set! xPitch (string->number (cadddr nameList)))
                            (set! yPitch (string->number (car (cddddr nameList))))
                            (axCreateContactArray cv netName viaNum 0 routeType
                              (list
                                (* pdefLengthResolution x0)
                                (* pdefLengthResolution y0)
                              )
                              xPitch yPitch xDuplicate yDuplicate '()
                            )
                          )
                        ) ; end if
                      ) ; end begin
                    ) ; end if
                  )
                )
                (else ; wire
                  (begin
                    (set! layerNum (car attr))
                    (set! width (dbFetchLayerInfo lib layerNum "defaultWidth"))
                    (set! x0 (caadr attr))
                    (set! y0 (cadadr attr))
                    (set! x1 (caaddr attr))
                    (set! y1 (cadr (caddr attr)))
                    (if (equal? first #t)
                      (begin
                        (set! pre_x x0)
                        (set! pre_y y0)
                        (set! first #f)
                      )
                      (begin
                        (if (equal? x0 '*)
                          (set! x0 pre_x)
                          (set! pre_x x0)
                        )
                        (if (equal? y0 '*)
                          (set! y0 pre_y)
                          (set! pre_y y0)
                        )
                      )
                    )
                    (if (equal? x1 '*)
                      (begin
                        (set! x1 x0)
                        (set! isVertical 1)
                      )
                      (set! pre_x x1)
                    )
                    (if (equal? y1 '*)
                      (begin
                        (set! y1 y0)
                        (set! isVertical 0)
                      )
                      (set! pre_y y1)
                    )
                    (if (= 0 isVertical)
                      (set! wireLength (abs (- x1 x0)))
                      (set! wireLength (abs (- y1 y0)))
                    )
                    (axCreateWire cv netName layerNum 0 2 width routeType
                      isVertical (* pdefLengthResolution wireLength)
                      (list
                        (* pdefLengthResolution x0)
                        (* pdefLengthResolution y0)
                      )
                    )
                  ) ; end begin
                ) ; end else
              )) ; end cond and lambda
              attrList
            ) ; end for-each
          )) ; end begin and lambda
          routeList
        ) ; end for-each
      )) ; end begin and lambda
      nlist
    ) ; end for-each
    #t
  )
))

(define findViaMaster (lambda (vList index)
  (let
    (
      (ret      0)
      (master '())
    )
    (call-with-current-continuation (lambda (ret)
      (for-each (lambda (v)
        (if (= index (caddr v))
          (begin
            (set! master v)
            (ret master)
          )
        ))
        vList
      )
    )) ; end call-with-current-continuation
    master
  )
))

(define string-parse (lambda (str)
  (let
    (
      (strList '())
      (numList '())
      (number  '())
      (len       0)
      (index     0)
    )
    (set! strList (string->list str))
    (set! len (length strList))
    (for-each (lambda (char)
      (cond
        ((equal? #\space char)
          (begin
            (if (not (null? number))
              (begin
                (if (equal? #f (string->number (list->string (reverse number))))
                  (set! number (list->string (reverse number)))
                  (set! number (string->number (list->string (reverse number))))
                )
                (set! numList (cons number numList))
                (set! number '())
              )
            )
            (set! index (+ index 1))
          )
        )
        ((= index (- len 1))
          (begin
            (set! number (cons char number))
            (if (equal? #f (string->number (list->string (reverse number))))
              (set! number (list->string (reverse number)))
              (set! number (string->number (list->string (reverse number))))
            )
            (set! numList (cons number numList))
          )
        )
        (else
          (begin
            (set! number (cons char number))
            (set! index (+ index 1))
          )
        )
      )) ; end cond and lambda
      strList
    ) ; end for-each
    (reverse numList)
  )
))

(define string-split-by-char (lambda (s c)
  (let
    (
      (result '())
      (sl (string-length s))
    )
    (if (zero? sl)
      result
      (let loop ((s s) (sp sl) (r result))
        (let ((nwp (string-skip-char s c sp)))
          (if (zero? nwp)
            r
            (let*
              ((nsp (string-skip-non-char s c nwp))
                (nr (cons (substring s nsp nwp) r)))
	      (if (zero? nsp)
                nr
                (loop s (- nsp 1) nr)
              )
            )
          )
        )
      )
    )
  )
))

(define string-skip-char (lambda (s c i)
  (cond
   ((zero? i) i)
   ((char=? c (string-ref s (- i 1))) (string-skip-char s c (- i 1)))
   (else i)
  )
))

(define string-skip-non-char (lambda (s c i)
  (cond
    ((zero? i) i)
    ((char=? c (string-ref s (- i 1))) i)
    (else (string-skip-non-char s c (- i 1)))
  )
))

